from __future__ import annotations

import sys
from collections.abc import Collection, Iterable
from typing import TYPE_CHECKING, Any, TypeVar, cast

import gymnasium as gym
import torch as th
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.distributions import \
    SquashedDiagGaussianDistribution  # StateDependentNoiseDistribution
from stable_baselines3.common.preprocessing import get_action_dim
from stable_baselines3.common.torch_layers import (CombinedExtractor,
                                                   FlattenExtractor)
from stable_baselines3.common.utils import get_schedule_fn
from stable_baselines3.sac import SAC
from stable_baselines3.sac.policies import Actor, BasePolicy, SACPolicy
from torch import nn

if TYPE_CHECKING:
    from stable_baselines3.common.torch_layers import BaseFeaturesExtractor
    from stable_baselines3.common.type_aliases import PyTorchObs, Schedule

# None = all, slice for boxes, collection for dict
InputSpec = slice | Collection[str] | None
# (action_dim, input)
NetArch = list[int] | dict[str, list[int]] | None
ActorSpec = tuple[int, InputSpec, NetArch]

T = TypeVar('T', bound=gym.Space[Any])


def make_subspace(space: T, inputs: InputSpec) -> T:
    if inputs is None:
        return space
    if isinstance(inputs, slice):
        if isinstance(space, gym.spaces.Box):
            return gym.spaces.Box(  # type: ignore
                space.low[inputs], space.high[inputs])
        else:
            raise ValueError("Slice indices must be used only with box spaces")
    if isinstance(space, gym.spaces.Dict):
        return gym.spaces.Dict({  # type: ignore
            k: v
            for k, v in space.spaces.items() if k in inputs
        })
    raise ValueError("String indices must be used only with dict spaces")


def make_actors(specs: list[ActorSpec], observation_space: gym.Space,
                action_space: gym.spaces.Box, **kwargs):
    actors = []
    out_index = 0
    for (dims, inputs, net_arch) in specs:
        actor_observation_space = make_subspace(observation_space, inputs)
        actor_action_space = make_subspace(action_space,
                                           slice(out_index, out_index + dims))
        out_index += dims
        actor_kwargs = dict(**kwargs)
        actor_kwargs['features_dim'] = gym.spaces.utils.flatdim(
            actor_observation_space)
        actor_kwargs['features_extractor'] = actor_kwargs[
            'features_extractor_class'](
                actor_observation_space,
                **actor_kwargs['features_extractor_kwargs'])
        del actor_kwargs['features_extractor_class']
        del actor_kwargs['features_extractor_kwargs']
        if net_arch is not None:
            actor_kwargs['net_arch'] = net_arch
        actors.append(
            Actor(actor_observation_space, actor_action_space, **actor_kwargs))
    return actors


def filter_obs(obs, spec: InputSpec):
    if spec is None:
        return obs
    if isinstance(spec, slice):
        return obs[..., spec]
    return {k: v for k, v in obs.items() if k in spec}


class ActorList(Actor):

    def __init__(self,
                 actor_specs: list[ActorSpec],
                 observation_space: gym.Space,
                 action_space: gym.spaces.Box,
                 normalize_images: bool = True,
                 **kwargs: Any):
        BasePolicy.__init__(self,
                            observation_space,
                            action_space,
                            normalize_images=normalize_images,
                            squash_output=True)
        self.actors = make_actors(actor_specs,
                                  observation_space=observation_space,
                                  action_space=action_space,
                                  normalize_images=normalize_images,
                                  **kwargs)
        for i, actor in enumerate(self.actors):
            self.add_module(f'actor_{i}', actor)
        action_dim = get_action_dim(self.action_space)
        self.action_dist = SquashedDiagGaussianDistribution(  # type: ignore[assignment]
            action_dim)
        self.filters = [f for _, f, _ in actor_specs]

    def set_training_mode(self, mode):
        for actor in self.actors:
            actor.set_training_mode(mode)

    def get_action_dist_params(
            self, obs: PyTorchObs
    ) -> tuple[th.Tensor, th.Tensor, dict[str, th.Tensor]]:
        outs: list[tuple[th.Tensor, th.Tensor, dict[str, th.Tensor]]] = [
            actor.get_action_dist_params(filter_obs(obs, f))
            for actor, f in zip(self.actors, self.filters, strict=True)
        ]
        return (th.concat([out[0] for out in outs],
                          dim=-1), th.concat([out[1] for out in outs],
                                             dim=-1), {})


def update_learning_rates(optimizer: th.optim.Optimizer,
                          learning_rates: Iterable[float]) -> None:
    for param_group, learning_rate in zip(optimizer.param_groups,
                                          learning_rates,
                                          strict=True):
        param_group["lr"] = learning_rate


class SplitSACPolicy(SACPolicy):
    """
    A SAC policy whose actor contains independent MLP
    that outputs some of the actions.

    That is, instead of the monolithic actor that takes
    all observations and computes all actions,
    the actor of this policy can use subset of the observations
    to compute some of the actions.

    The sub-modules are configured from argument :py:obj:`actor_specs` in
    ``policy_kwargs`` of type :py:type:`list[ActorSpec]`.

    The different tuples ``(action_size, input_spec, net_arch)``
    each configure one of the MLPs that computes ``action_size`` actions
    using observations specified by``input_spec`` with an architecture ``net_arch``.

    The list should be ordered and actions sizes
    should sum up to the total size of the action space.

    For example:

    >>> env.observation_space
    Dict('a': Box(0.0, 1.0, (1,), float32), 'b': Box(0.0, 1.0, (1,), float32))
    >>> env.action_space
    Box(-1.0, 1.0, (2,), float32)
    >>> actor_specs = [(1, None, [64, 64]), (1, ['a'], [16, 16])]
    >>> model = SAC(env, policy_kwargs={'actor_specs': actor_specs})

    creates a model with a policy that computes two actions:

    - the first using a 64 + 64 MLP from ``a`` and ``b``
    - the second using a 16 + 16 MLP solely from ``a``

    """

    actor: ActorList
    actor_specs: list[ActorSpec]

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.spaces.Box,
        lr_schedule: Schedule,
        net_arch: list[int] | dict[str, list[int]] | None = None,
        activation_fn: type[nn.Module] = nn.ReLU,
        use_sde: bool = False,
        log_std_init: float = -3,
        use_expln: bool = False,
        clip_mean: float = 2.0,
        features_extractor_class: type[BaseFeaturesExtractor] | None = None,
        features_extractor_kwargs: dict[str, Any] | None = None,
        normalize_images: bool = True,
        optimizer_class: type[th.optim.Optimizer] = th.optim.Adam,
        optimizer_kwargs: dict[str, Any] | None = None,
        n_critics: int = 2,
        share_features_extractor: bool = False,
        actor_specs: list[ActorSpec] = [],
    ):
        """
        Constructs a new instance.

        Same parameters of the super class :py:class:`SACPolicy` constructor,
        with the addition of :py:obj:`actor_specs`.

        :param      actor_specs:  The actor specs.
        """

        if features_extractor_class is None:
            if isinstance(observation_space, gym.spaces.Box):
                features_extractor_class = FlattenExtractor
            else:
                features_extractor_class = CombinedExtractor
        self.actor_specs = actor_specs

        SACPolicy.__init__(
            self,
            observation_space,
            action_space,
            lr_schedule,
            net_arch,
            activation_fn,
            use_sde,
            log_std_init,
            use_expln,
            clip_mean,
            features_extractor_class,
            features_extractor_kwargs,
            normalize_images,
            optimizer_class,
            optimizer_kwargs,
            n_critics,
            share_features_extractor,
        )

    def _build(self, lr_schedule: Schedule) -> None:
        super()._build(lr_schedule)
        self.actor.optimizer = self.optimizer_class(
            [{
                'params': actor.parameters()
            } for actor in self.actor.actors],
            lr=lr_schedule(1),  # type: ignore[call-arg]
            **self.optimizer_kwargs)

    def make_actor(
            self,
            features_extractor: BaseFeaturesExtractor | None = None) -> Actor:
        actor_kwargs = dict(**self.actor_kwargs)
        actor_kwargs[
            'features_extractor_class'] = self.features_extractor_class
        actor_kwargs[
            'features_extractor_kwargs'] = self.features_extractor_kwargs
        return ActorList(actor_specs=self.actor_specs,
                         **actor_kwargs).to(self.device)  # type: ignore


def setup_actor_learning_rates(model: SAC,
                               learning_rates: list[Schedule | float]) -> None:
    """
    Set the learning rates of the sub-modules composing the actor.

    Only applied if the model policy is of class :py:class:`SplitSACPolicy`.

    :param      model:           The model

    :param   learning_rates:  A list of tuples
        ``(number of training steps, learning rate)``,
        one for each sub-module of the actor.
    """

    if isinstance(model.policy, SplitSACPolicy):
        print("Model policy type is not SplitSACPolicy", file=sys.stderr)
        return

    lr_schedules = [get_schedule_fn(lr) for lr in learning_rates]

    def update_learning_rate(
            optimizers: list[th.optim.Optimizer] | th.optim.Optimizer) -> None:
        lrs = [fn(model._current_progress_remaining) for fn in lr_schedules]
        for i, lr in enumerate(lrs):
            model.logger.record(f"train/learning_rate_actor{i}", lr)
        update_learning_rates(model.actor.optimizer, lrs)
        if isinstance(optimizers, list):
            SAC._update_learning_rate(model, optimizers[1:])

    model._update_learning_rate = update_learning_rate  # type: ignore[method-assign]


class AlternateActorCallback(BaseCallback):
    """
    A callback that alternates training modules that
    compose the actor of :py:class:`SplitSACPolicy`.
    """

    def __init__(self,
                 learning_rates: list[tuple[int, Schedule | float]],
                 exclusive: bool = True,
                 verbose: int = 0):
        """
        Constructs a new instance.

        :param      learning_rates:  The learning rates, as a list of
            tuple (number of training steps, learning rate),
            one for each sub-module of the :py:class:`SplitSACPolicy`
            actor.
        :param      exclusive: Whether to set the learning rate to the
                other modules (``exclusive=False) or to remove their
                parameters from the optimization (``exclusive=True)

        :param      verbose:         The verbosity level
        """
        super().__init__(verbose)
        self._next_actor_index = 0
        self._next_actor_step = 0
        self._learning_rates = learning_rates
        self._exclusive = exclusive

    def _on_step(self) -> bool:
        if self.num_timesteps >= self._next_actor_step:
            # print(self.model.policy.actor.optimizer)
            steps, lr = self._learning_rates[self._next_actor_index]
            self.logger.record("train/actor_index", self._next_actor_index)
            self._next_actor_step = self.num_timesteps + steps
            if self._exclusive:
                actor = cast('ActorList', self.model.policy.actor)
                actor.optimizer = self.model.policy.optimizer_class(
                    actor.actors[self._next_actor_index].parameters(),
                    lr=lr,  # type: ignore[call-arg]
                    **self.model.policy.optimizer_kwargs)
                self.model.lr_schedule = get_schedule_fn(lr)
            else:
                learning_rates = [
                    lr if i == self._next_actor_index else 0
                    for i, (_, lr) in enumerate(self._learning_rates)
                ]
                setup_actor_learning_rates(cast('SAC', self.model),
                                           learning_rates)
            self._next_actor_index = (self._next_actor_index + 1) % len(
                self._learning_rates)
        return True



from __future__ import annotations

import dataclasses as dc
import warnings
from types import TracebackType
from typing import Any, cast

import gymnasium as gym
import numpy as np
from navground import core

from ..types import Array
from .base import ActionConfig, DataclassConfig


def to_py(value: Array) -> Any:
    try:
        return value.item()
    except ValueError:
        pass
    return value.tolist()


class TemporaryBehaviorParams:

    def __init__(self, behavior: core.Behavior, params: dict[str, Any]):
        super().__init__()
        self._behavior = behavior
        self._params = params
        self._old_params: dict[str, Any] = {}

    def __enter__(self) -> None:
        for k, v in self._params.items():
            self._old_params[k] = getattr(self._behavior, k)
            setattr(self._behavior, k, v)

    def __exit__(self, exc_type: type[BaseException] | None,
                 exc_value: BaseException | None,
                 exc_tb: TracebackType | None) -> None:
        for k, v in self._old_params.items():
            setattr(self._behavior, k, v)


def single_param_space(low: Any = 0,
                       high: Any = 1,
                       dtype: str = 'float',
                       discrete: bool = False,
                       normalized: bool = False) -> gym.Space[Any]:
    if discrete:
        return gym.spaces.Discrete(start=low, n=int(high - low))
    if normalized:
        low = -1
        high = 1
    box_type = np.dtype(dtype).type
    return gym.spaces.Box(low, high, dtype=box_type)


def param_space(params: dict[str, dict[str, Any]],
                normalized: bool = False) -> gym.spaces.Dict:
    return gym.spaces.Dict({
        key:
        single_param_space(normalized=normalized, **value)
        for key, value in params.items()
    })


@dc.dataclass(repr=False)
class ModulationActionConfig(DataclassConfig, ActionConfig, register_name="Modulation"):
    """
    Configuration of the conversion between modulation actions
    and control commands.

    :param params: The parameters to modulate.

    Actions are values of parameters of a behavior.

    """

    params: dict[str, dict[str, Any]] = dc.field(default_factory=dict)
    """The parameters to modulate as a dictionary name -> space configuration"""

    def __post_init__(self) -> None:
        self.param_space = param_space(self.params, normalized=True)

    def is_configured(self, warn: bool = False) -> bool:
        if not self.params:
            warnings.warn("No parameters", stacklevel=1)
            return False
        return True

    def normalize(self, key: str, value: Any) -> Any:
        param = self.params[key]
        if 'discrete' in param:
            return value
        low = param['low']
        high = param['high']
        return np.clip(-1, 1, -1 + 2 * (value - low) / (high - low))

    def de_normalize(self, key: str, value: Any) -> Any:
        param = self.params[key]
        if 'discrete' in param:
            return value
        low = param['low']
        high = param['high']
        return low + (value + 1) / 2 * (high - low)

    @property
    def space(self) -> gym.Space[Any]:
        """
        The action space is a flattened :py:class:`gymnasium.spaces.Dict`
        with one entry per controlled parameter.
        """
        if self.param_space:
            return gym.spaces.flatten_space(self.param_space)
        else:
            return self.param_space

    def get_params_from_action(self, action: Array) -> dict[str, Any]:
        values: dict[str,
                     Array] = gym.spaces.unflatten(self.param_space, action)
        return {k: to_py(self.de_normalize(k, v)) for k, v in values.items()}

    def get_cmd_from_action(self, action: Array,
                            behavior: core.Behavior | None,
                            time_step: float) -> core.Twist2:
        assert behavior is not None
        params = self.get_params_from_action(action)
        with TemporaryBehaviorParams(behavior, params):
            cmd = behavior.compute_cmd(time_step)
        return cmd

    def get_action(self, behavior: core.Behavior, time_step: float) -> Array:
        params = {
            k: self.normalize(k, getattr(behavior, k))
            for k in self.param_space
        }
        return cast('Array', gym.spaces.flatten(self.param_space, params))

    def configure(self, behavior: core.Behavior) -> None:
        pass

    @property
    def asdict(self) -> dict[str, Any]:
        rs = dc.asdict(self)
        rs['type'] = self.__class__.__name__
        return rs



from __future__ import annotations

from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from pettingzoo.utils.env import ParallelEnv
    from stable_baselines3.common.vec_env import VecEnv

    from ..types import Action, Observation

import types


def make_vec_from_penv(env: ParallelEnv[int, Observation, Action],
                       num_envs: int = 1,
                       processes: int = 1,
                       seed: int = 1) -> VecEnv:
    """
    Creates an imitation-compatible vectorized enviroment from
    a :py:class:`PettingZoo  Parallel environment <pettingzoo.utils.env.ParallelEnv>`,
    similarly as :py:func:`navground.learning.parallel_env.make_vec_from_penv` but
    adding a (custom) ``RolloutInfoWrapper`` to collect rollouts.

    It first creates a :py:class:`supersuit.vector.MarkovVectorEnv`, and then
    applies :py:func:`supersuit.concat_vec_envs_v1` to concatenate ``number`` copies of it.

    :param      env:       The environment
    :param      num_envs:  The number of pettingzoo environments to stuck together
    :param      processes: The number of (parallel) processes
    :param      seed:      The seed

    :returns:   The vectorized environment.
    """
    import supersuit  # type: ignore[import-untyped]

    from .parallel_rollout_wrapper import RolloutInfoWrapper

    env = RolloutInfoWrapper(env)
    menv = supersuit.vector.MarkovVectorEnv(env, black_death=True)
    venv = supersuit.concat_vec_envs_v1(menv,
                                        num_envs,
                                        num_cpus=processes,
                                        base_class="stable_baselines3")

    def get_attr(self, attr_name: str, indices=None) -> list:
        if indices is None:
            indices = range(num_envs * menv.num_envs)
        if isinstance(indices, int):
            indices = [indices]
        rs = []
        for i in indices:
            penv = self.vec_envs[i // menv.num_envs].par_env
            rs.append(getattr(penv, attr_name))
        if attr_name == 'render_mode':
            # This is an attribute of vector environments themselves
            return rs[0]
        return rs

    def set_attr(self, attr_name: str, value, indices=None) -> None:
        if indices is None:
            indices = range(num_envs * menv.num_envs)
        if isinstance(indices, int):
            indices = [indices]
        for i in indices:
            penv = self.vec_envs[i // menv.num_envs].par_env
            setattr(penv, attr_name, value)

    cenv = venv.unwrapped
    cenv.get_attr = types.MethodType(get_attr, cenv)
    cenv.set_attr = types.MethodType(set_attr, cenv)

    for i, env in enumerate(venv.venv.vec_envs):
        env.reset(seed + i)
    return cast("VecEnv", venv)



from __future__ import annotations

import numpy as np
from navground import core, sim


class CommSensor(sim.Sensor, name="Comm", include_properties_of=['Sensor']):
    """
    A sensor that receive messages broadcasted by other agents.
    that it stores in field "comm".

    :param      binarize:  Whether to binarize the received message
    :param      size:      The message size
    :param      name:      The namespace
    """

    def __init__(self, binarize: bool = True, size: int = 1, name: str = ''):
        super().__init__(name=name)
        self._binarize = binarize
        self._size = size

    @property
    @sim.register(True, "Whether to binarize comm")
    def binarize(self) -> bool:
        return self._binarize

    @binarize.setter
    def binarize(self, value: bool) -> None:
        self._binarize = value

    @property
    @sim.register(1, "size")
    def size(self) -> int:
        return self._size

    @size.setter
    def size(self, value: int) -> None:
        self._size = max(0, value)

    def update(self, agent: sim.Agent, world: sim.World,
               state: core.EnvironmentState) -> None:
        if not isinstance(state, core.SensingState):
            return
        cs = []
        for other in world.agents:
            if other is not agent:
                if other.behavior and hasattr(other.behavior, "_comm"):
                    cs.append(
                        other.behavior._comm)  # type: ignore[attr-defined]
                else:
                    cs.append(np.zeros(self.size))
        data = np.concatenate(cs, dtype=np.float32)
        if self._binarize:
            data = (data > 0).astype(dtype=int)
        self.get_or_init_buffer(state, "comm").data = data

    def get_description(self) -> dict[str, core.BufferDescription]:
        if self.binarize:
            desc = core.BufferDescription([self.size], np.dtype(np.int8), 0, 1)
        else:
            # CHANGED low
            desc = core.BufferDescription([self.size], np.float32, -1, 1)
        return {self.get_field_name("comm"): desc}



from typing import Any

import gymnasium as gym
from navground import sim

from ..config import ControlActionConfig, DefaultObservationConfig
from ..env import BaseEnv
from ..parallel_env import BaseParallelEnv, shared_parallel_env
from ..rewards import SocialReward


def get_env(flat: bool = True,
            use_acceleration_action: bool = True,
            multi_agent: bool = False,
            **kwargs: Any) -> BaseEnv | BaseParallelEnv:
    """
    Creates the an environment where 20 agents travel back and forth
    between way-points, crossing in the middle.

    :param flat: Whether the observation space is flat
    :param use_acceleration_action: Whether actions are acceleration or velocities
    :param multi_agent: Whether to expose all agents or just one.
    :param kwargs: Arguments passed to the environment constructor

    :returns: A Parallel PettingZoo environment if `multi_agent` is set,
        else a Gymnasium environment.

    """

    scenario = sim.load_scenario("""
type: Cross
agent_margin: 0.1
side: 4
target_margin: 0.1
tolerance: 0.5
groups:
  -
    type: thymio
    number: 20
    radius: 0.1
    control_period: 0.1
    speed_tolerance: 0.02
    color: gray
    kinematics:
      type: 2WDiff
      wheel_axis: 0.094
      max_speed: 0.12
    behavior:
      type: HL
      optimal_speed: 0.12
      horizon: 5.0
      tau: 0.25
      eta: 0.5
      safety_margin: 0.1
    state_estimation:
      type: Bounded
      range: 5.0
""")

    sensor = sim.load_state_estimation("""
type: Discs
number: 5
range: 5.0
max_speed: 0.12
max_radius: 0.1
""")

    observation_config = DefaultObservationConfig(
        flat=flat, include_target_direction=True, include_target_distance=True)
    action_config = ControlActionConfig(
        use_acceleration_action=use_acceleration_action)
    if use_acceleration_action:
        observation_config.include_angular_speed = True
        observation_config.include_velocity = True
        action_config.max_acceleration = 1.0
        action_config.max_angular_acceleration = 10.0
    vs: dict[str, Any] = dict(scenario=scenario,
                              sensor=sensor,
                              action=action_config,
                              observation=observation_config,
                              reward=SocialReward(),
                              time_step=0.1)
    vs.update(kwargs)
    if not multi_agent:
        return gym.make('navground', **vs)
    return shared_parallel_env(**vs)



from __future__ import annotations

from typing import Any

import numpy as np
from torch import no_grad
from torchrl.envs.utils import (ExplorationType,  # type: ignore
                                set_exploration_type)


def evaluate_policy(
    policy: Any,
    env: Any,
    n_eval_episodes: int = 10,
    deterministic: bool = True,
    return_episode_rewards: bool = False,
    warn: bool = True,
) -> tuple[float, float] | tuple[list[float], list[int]]:
    """
    Similar interface as
    StableBaseline3 :py:func:`stable_baselines3.common.evaluation.evaluate_policy`
    but for TorchRL environments (and policies).

    :param      policy:                  The policy
    :param      env:                     The environment
    :param      n_eval_episodes:         The number of episodes
    :param      deterministic:           Whether to evaluate the policy deterministic
    :param      return_episode_rewards:  Whether to return individual episode rewards
                                         (vs aggregate them)
    :param      warn:                    Whether to emit warnings

    :returns:   If ``return_episode_rewards`` is set,
        a tuple (list of cumulated episodes rewards', list of episodes length)
        else, a tuple (average episodes rewards, std dev of episodes rewards)
    """
    num = len(env.env.possible_agents)
    episode_rewards = []
    episode_lengths = []
    for _ in range(n_eval_episodes // num):
        with set_exploration_type(ExplorationType.DETERMINISTIC), no_grad():
            rollout = env.rollout(max_steps=10_000, policy=policy)
        reward = rollout['next']['agent']['reward'].numpy()
        episode_rewards.append(np.sum(reward, axis=0).flatten())
        length = len(reward)
        episode_lengths.append(length)
        episode_lengths.append(length)
    episode_rewards = np.concatenate(episode_rewards)
    mean_reward = np.mean(episode_rewards)
    std_reward = np.std(episode_rewards)
    if return_episode_rewards:
        return episode_rewards, episode_lengths
    return mean_reward, std_reward



from __future__ import annotations

from typing import cast

from navground import core

from ...config import ControlActionConfig, DefaultObservationConfig
from ...parallel_env.join import stack_observations
from ...types import AnyPolicyPredictor, ObservationTransform, PathLike
from .base import BasePolicyMixin


class GroupPolicyBehavior(core.BehaviorGroup):

    def __init__(self, policy: AnyPolicyPredictor):
        super().__init__()
        self.policy = policy

    def compute_cmds(self, time_step: float) -> list[core.Twist2]:

        behaviors = cast('list[GroupedPolicyBehavior]', self.members)

        obs = stack_observations({
            i: behavior.update_observation()
            for i, behavior in enumerate(behaviors)
        })
        acts, _ = self.policy.predict(obs, deterministic=True)
        acts = acts.reshape(len(self.members), -1)
        return [
            behavior.get_cmd_from_action(act, time_step)
            for behavior, act in zip(behaviors, acts, strict=True)
        ]


class GroupedPolicyBehavior(BasePolicyMixin,
                            core.BehaviorGroupMember,
                            name="GroupPolicy",
                            include_properties_of=[BasePolicyMixin]):

    _groups: dict[int, core.BehaviorGroup] = {}

    def __init__(self,
                 kinematics: core.Kinematics | None = None,
                 radius: float = 0.0,
                 policy: AnyPolicyPredictor | None = None,
                 policy_path: PathLike = '',
                 action_config: ControlActionConfig = ControlActionConfig(),
                 observation_config:
                 DefaultObservationConfig = DefaultObservationConfig(),
                 deterministic: bool = False,
                 pre: ObservationTransform | None = None):
        BasePolicyMixin.__init__(self, policy, policy_path, action_config,
                                 observation_config, deterministic, pre)
        core.BehaviorGroupMember.__init__(self, kinematics, radius)

    def make_group(self) -> core.BehaviorGroup:
        if self.policy_path:
            policy: AnyPolicyPredictor | None = self.load_policy(self.policy_path)
        else:
            policy = self._policy
        if not policy:
            raise RuntimeError("Policy not set")
        return GroupPolicyBehavior(policy)

    def get_groups(self) -> dict[int, core.BehaviorGroup]:
        return self._groups

    def get_group_hash(self) -> int:
        if self.policy_path:
            return abs(hash(self.policy_path))
        return abs(hash(self._policy))

    def close(self) -> None:
        core.BehaviorGroupMember.close(self)

    def prepare(self) -> None:
        BasePolicyMixin.prepare(self)
        core.BehaviorGroupMember.prepare(self)



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING

from imitation.algorithms import bc
from imitation.data import rollout as rollout_without_info

from . import rollout

if TYPE_CHECKING:
    import numpy as np
    from stable_baselines3.common import policies


class RolloutStatsComputer(bc.RolloutStatsComputer):

    def __call__(
        self,
        policy: policies.ActorCriticPolicy,
        rng: np.random.Generator,
    ) -> Mapping[str, float]:
        if self.venv is not None and self.n_episodes > 0:
            trajs = rollout.generate_trajectories(
                policy,
                self.venv,
                rollout_without_info.make_min_episodes(self.n_episodes),
                rng=rng,
            )
            return rollout_without_info.rollout_stats(trajs)
        else:
            return dict()


# TODO(Jerome): monkey patching for now
# In the future, better to subclass BC.
bc.RolloutStatsComputer = RolloutStatsComputer  # type: ignore[misc]

BC = bc.BC

__all__ = ['BC', 'RolloutStatsComputer']



from __future__ import annotations

import dataclasses as dc
from typing import Any, cast

import gymnasium as gym
import numpy as np
from navground import core

from ..types import Action
from .control_action import ControlActionConfig
from .discrete_control_action import to_continuous, to_discrete


def to_multibinary(value: int, number: int) -> list[int]:
    rs = []
    for _ in range(number):
        rs.append(value % 2)
        value = value // 2
    return rs


@dc.dataclass
class DiscreteControlActionWithCommConfig(
        ControlActionConfig, register_name="DiscreteControlWithComm"):
    """
    Actions are composed of control values and the message to be transmitted

    Message {0, 1} and controls {0, 1, 2} are discrete and
    encoded using a mixed numerical scheme:
    - the lower part with a ternary encoding for controls
    - the upper part with binary encoding for messages.

    Action to be actuated are decomposed:
    - the control is converted to a :py:class:`navground.core.Twist2`,
    - the message is set as attribute ``Behavior._comm``.

    :param dtype: The data type

    :param dof: The number of degrees of freedom of the agent

    :param max_speed: The upper bound of the speed.

    :param max_angular_speed: The upper bound of the angular speed.

    :param max_acceleration: The upper bound of the acceleration.

    :param max_angular_acceleration: The upper bound of the angular acceleration.

    :param use_acceleration_action: Whether actions are accelerations.
                                    If not set, actions are velocities.

    :param use_wheels: Whether action uses wheel speeds/acceleration
                       instead of body speeds/acceleration.
                       Only effective if the b behavior has a wheeled kinematics.

    :param fix_orientation: Whether to force the agent not to control orientation,
                            i.e., to not include the angular command in actions.

    :param has_wheels: Whether the agent as wheels.
                       If None, it will defer to the agent kinematics.

    :param comm_size: The size of the message to broadcast

    """
    comm_size: int = 1
    """The size of the message to broadcast"""

    @property
    def comm_space(self) -> gym.spaces.Box:
        """
        The box space of the broadcasted messages with dimension
        :py:meth:`ControlActionConfig.action_size` and values in {0, 1}.

        :returns: the box space
        """
        return gym.spaces.Box(low=0,
                              high=1,
                              shape=(self.comm_size, ),
                              dtype=np.int8)

    def configure(self, behavior: core.Behavior) -> None:
        super().configure(behavior)
        self._cont_space = cast('gym.spaces.Box', super().space)

    def get_cmd_from_action(self, action: Action,
                            behavior: core.Behavior | None,
                            time_step: float) -> core.Twist2:
        n = 3**self.action_size
        comm = to_multibinary(int(action) // n, self.comm_size)
        if behavior:
            behavior._comm = np.array(comm, dtype=np.int8)  # type: ignore[attr-defined]
        action = int(action) % n
        act = np.array(to_continuous(int(action), self.action_size),
                       dtype=self._cont_space.dtype)
        act = self._cont_space.low + act * (self._cont_space.high -
                                            self._cont_space.low)
        return super().get_cmd_from_action(act, behavior, time_step)

    def get_action(self, behavior: core.Behavior, time_step: float) -> Action:
        act = ControlActionConfig.get_action(self, behavior, time_step)
        if hasattr(behavior, '_comm'):
            comm = behavior._comm  # type: ignore[attr-defined]
        else:
            comm = []
        if len(comm) < self.comm_size:
            comm = comm + [0] * (self.comm_size - len(comm))
        th = (self._cont_space.high + self._cont_space.low) / 2
        dact = to_discrete(act, threshold=th)
        dcomm = sum(v * (2**i) for i, v in enumerate(comm))
        d = dact + dcomm * (3**self.action_size)
        return np.array([d], dtype=np.int64)

    @property
    def space(self) -> gym.Space[Any]:
        """
        The control space {0 (backwards), 1 (do-nothing), 2 (forwards)}
        multiplied by the communication space {0, 1}
        """
        return gym.spaces.Discrete(3**self.action_size * 2**self.comm_size)



from __future__ import annotations

from navground import core


def compute_speeds(values: list[tuple[float, float]], b) -> list[float]:
    # all have already passed the pad (half-way) or one has already exited the pad
    # => proceed at full speed
    assert len(values) == 2
    ss = [s for _, s in values]
    if all(x > 0 for x, _ in values) or any(x > b for x, _ in values):
        return ss
    # both are in the first half of the pad.
    # if the fastest to exit is moving backwards, let's do it
    if all(x > -b for x, _ in values):
        time_to_exit_fwd = [(b - x) / s for x, s in values]
        time_to_exit_bwd = [(b + x) / s for x, s in values]
        if min(time_to_exit_fwd) > min(time_to_exit_bwd):
            i = 0 if time_to_exit_bwd[0] < time_to_exit_bwd[1] else 1
            ss[i] *= -1
        return ss
    time_to_exit = [(b - x) / s for x, s in values]
    i = 0 if time_to_exit[0] > time_to_exit[1] else 1
    ss[i] = min(ss[i], (-b - values[i][0]) / time_to_exit[1 - i])
    return ss


class PadGroupBehavior(core.BehaviorGroup):

    def __init__(self, pad_width: float = 0.5):
        super().__init__()
        self.pad_width = 0.5

    def compute_cmds(self, time_step: float) -> list[core.Twist2]:
        assert len(self.members) == 2 and all(
            behavior.target.direction is not None for behavior in self.members)
        vs = [
            (
                behavior.target.direction.dot(  # type: ignore[union-attr]
                    behavior.position),
                behavior.get_target_speed()) for behavior in self.members
        ]
        speeds = compute_speeds(vs, self.pad_width / 2)
        return [
            core.Twist2((speed, 0), 0, frame=core.Frame.relative)
            for speed in speeds
        ]


class CentralizedPadBehavior(core.BehaviorGroupMember, name="CentralizedPad"):
    """
    A centralized behavior that applies a (quasi) optimal strategy
    in :py:class:`navground.learning.scenarios.PadScenario`:

    - if both agents have already passed the mid-point of the pad,
      or at least one agent has already passed the pad:
      they advance at full speed
    - if both agents are in the first half of the pad:
      if the fastest way for one agent to exit first from the path
      is to move backwards:
      the agent move forward at full speed, while the
      other keeps advancing at full speed
    - else:
      the agent predicted to exit the pad second, adjust its speed like

      .. code-block::

         speed = min(optimal_speed,
                     distance_to_enter_pad / other_agent_time_to_exit_the_pad)``

      where the computation uses the other agent *optimal speed* to predict:

      .. code-block::

         other_agent_time_to_exit_the_pad = (
             other_agent_distance_to_exit_pad / other_agent_optional_speed)``

    """

    _groups: dict[int, core.BehaviorGroup] = {}

    def __init__(self,
                 kinematics: core.Kinematics | None = None,
                 radius: float = 0,
                 pad_width: float = 0.5):
        super().__init__(kinematics, radius)

        self._pad_width = pad_width

    @property
    @core.register(0.5, "Pad width")
    def pad_width(self) -> float:
        return self._pad_width

    @pad_width.setter
    def pad_width(self, value: float) -> None:
        self._pad_width = max(0, value)

    def make_group(self) -> core.BehaviorGroup:
        return PadGroupBehavior(pad_width=self.pad_width)

    def get_groups(self) -> dict[int, core.BehaviorGroup]:
        return self._groups



def test_import():
    import navground.learning.env  # noqa


def test_create_env():
    from navground.learning.examples.corridor_with_obstacle import get_env

    env = get_env()
    env.reset(seed=0)


def test_load_env():
    from navground.learning.io import load_env
    import pathlib

    path = pathlib.Path(__file__).parent / 'env.yaml'
    env = load_env(path)
    env.reset(seed=0)



from __future__ import annotations

import dataclasses as dc
from collections.abc import Collection, Mapping
from typing import Any, cast

import yaml
from navground import sim

from ..indices import Indices
from ..types import (AnyPolicyPredictor, JSONAble, PathLike, Reward,
                     SensorLike, SensorSequenceLike, TerminationCondition)
from .base import ActionConfig, ObservationConfig


def load_sensor(value: str) -> sim.Sensor | None:
    se = sim.load_state_estimation(value)
    if isinstance(se, sim.Sensor):
        return se
    print(f"Invalid sensor {se}")
    return None


def load_sensors(value: str) -> list[sim.Sensor]:
    ds = yaml.safe_load(value)
    ses = (sim.load_sensor(yaml.dump(d)) for d in ds)
    return [se for se in ses if se]


def make_sensor(value: SensorLike | None) -> sim.Sensor | None:
    if isinstance(value, dict):
        value = yaml.dump(value)
    if isinstance(value, str):
        value = load_sensor(value)
    return value


def make_sensors(value: SensorSequenceLike) -> list[sim.Sensor]:
    if isinstance(value, str):
        return load_sensors(value)
    return [x for x in (make_sensor(se) for se in value) if x]


def get_sensor_as_dict(sensor: SensorLike | None) -> dict[str, Any]:
    if sensor is None:
        return {}
    if isinstance(sensor, dict):
        return sensor
    if isinstance(sensor, sim.Sensor):
        sensor = sim.dump(sensor)
    return cast("dict[str, Any]", yaml.safe_load(sensor))


@dc.dataclass
class GroupConfig:
    """
    Configure a group of navground agents so that they
    are exposed in a :py:class:`gymnasium.Env` (only single agents)
    or :py:class:`pettingzoo.utils.env.ParallelEnv` (one or more agents).

    Group configurations can be stuck. For instance, the configuration provided
    to define the environment for training needs needs to define at least
    :py:attr:`indices`, :py:attr:`action` and
    :py:attr:`observation`, and possibly also one :py:attr:`sensors`.
    Instead, for evaluation, we may provide just the policy,
    as the agents are already configured.

    :param indices: The indices of the agents
                    in the :py:attr:`navground.sim.World.agents` list.

    :param action: The actions configuration.

    :param observation: The observations configuration.

    :param sensor: An optional sensor that will be added to :py:obj:`sensors`.

    :param sensors: A sequence of sensor to generate observations for the agents
                    or its YAML representation. If
                    Items of class :py:class:`str` will be interpreted as the YAML
                    representation of a sensor.
                    Items of class :py:class:`dict` will be dumped to YAML and
                    then treated as a :py:class:`str`.
                    If empty, it will use the agents' own sensors.

    :param reward: An optional reward function to use.

    :param color : An optional color for the agents (only used for displaying)

    :param tag : An optional tag added to the agents :py:attr:`navground.sim.Agent.tags`,
                 (does not impact simulation, only used to identify the agents)

    :param policy: The policy assigned to the agents during evaluation.

    :param deterministic: Whether the agents apply such policy deterministically.

    :param terminate_on_success: Whether to terminate when the agent succeeds

    :param terminate_on_failure: Whether to terminate when the agent fails

    :param success_condition: Optional success condition

    :param failure_condition: Optional failure condition
    """
    indices: Indices = Indices.all()
    """The indices of the agents in the :py:attr:`navground.sim.World.agents` list"""
    action: ActionConfig | None = None
    """The actions configuration"""
    observation: ObservationConfig | None = None
    """The observations configuration"""
    sensor: dc.InitVar[SensorLike | None] = None
    sensors: SensorSequenceLike = dc.field(default_factory=list)
    """List of sensors to generate observations for the agents"""
    reward: Reward | None = None
    """The reward function"""
    color: str = ''
    """The agents' color"""
    tag: str = ''
    """An optional additional tag for the agents"""
    policy: AnyPolicyPredictor | PathLike = ''
    """The policy assigned to the agents (during evaluation)"""
    deterministic: bool = False
    """Whether the agents apply such policy deterministically"""
    terminate_on_success: bool | None = None
    """Whether to terminate when the agent succeeds"""
    terminate_on_failure: bool | None = None
    """Whether to terminate when the agent fails"""
    success_condition: TerminationCondition | None = None
    """Success condition"""
    failure_condition: TerminationCondition | None = None
    """Failure condition"""

    def __post_init__(self, sensor: SensorLike | None) -> None:
        self.indices = Indices(self.indices)
        self.sensors = make_sensors(self.sensors)
        sensor = make_sensor(sensor)
        if sensor:
            if not isinstance(self.sensors, list):
                self.sensors = []
            self.sensors.append(sensor)

    @property
    def asdict(self) -> dict[str, Any]:
        """
        A JSON-able representation of the configuration

        :returns:  A JSON-able dict
        """
        rs: dict[str, Any] = {}
        if self.color:
            rs['color'] = self.color
        if self.tag:
            rs['tag'] = self.tag
        if self.deterministic:
            rs['deterministic'] = self.deterministic
        if self.observation:
            rs['observation'] = self.observation.asdict
        if self.action:
            rs['action'] = self.action.asdict
        if self.reward:
            rs['reward'] = self.reward.asdict
        if self.indices:
            rs['indices'] = self.indices.asdict
        rs['terminate_on_success'] = self.terminate_on_success
        rs['terminate_on_failure'] = self.terminate_on_failure
        rs['sensors'] = [get_sensor_as_dict(sensor) for sensor in self.sensors]
        return rs

    @classmethod
    def from_dict(cls, value: Mapping[str, Any]) -> GroupConfig:
        loadables: dict[str, type] = {
            'observation': ObservationConfig,
            'action': ActionConfig,
            'reward': Reward,
            'indices': Indices,
        }
        kwargs: dict[str, Any] = {
            k: value[k]
            for k in ('color', 'tag', 'deterministic', 'sensors', 'sensor',
                      'terminate_on_success', 'terminate_on_failure')
            if k in value
        }
        kwargs.update((k, cast("JSONAble", item_cls).from_dict(value[k]))
                      for k, item_cls in loadables.items() if k in value)
        return GroupConfig(**kwargs)

    def get_sensors(self) -> list[sim.Sensor]:
        return make_sensors(self.sensors)


def get_first_reward(groups: Collection[GroupConfig]) -> Reward | None:
    for group in groups:
        if group.reward:
            return group.reward
    return None


def merge_groups_configs(groups: Collection[GroupConfig],
                         other_groups: Collection[GroupConfig],
                         max_number: int) -> list[GroupConfig]:
    """
    Overlaps two sets of group configurations, letting the first
    inherit attributes from the second if they are not specified.

    :param      groups:        A collection of groups configuration
    :param      other_groups:  Another collection of groups configuration to be inherited from.
    :param      max_number:    The maximum number of agents, needed to intersect
                               the groups :py:attr:`GroupConfig.indices` to identify
                               which agents belong to groups in both sets.
    :returns:   A new list of groups, where the attributes from both groups collections.
    """
    gs: list[GroupConfig] = []
    for g1 in groups:
        for g2 in other_groups:
            indices = g1.indices.intersect(g2.indices, max_number)
            if indices:
                g = dc.replace(g1, indices=indices)
                if g1.action is None:
                    g.action = g2.action
                if not g1.sensors:
                    g.sensors = g2.sensors
                if g1.reward is None:
                    g.reward = g2.reward
                if g1.observation is None:
                    g.observation = g2.observation
                if g1.color is None:
                    g.color = g2.color
                if g1.tag is None:
                    g.tag = g2.tag
                if g1.policy is None:
                    g.policy = g2.policy
                if g1.terminate_on_success is None:
                    g.terminate_on_success = g2.terminate_on_success
                if g1.terminate_on_failure is None:
                    g.terminate_on_failure = g2.terminate_on_failure
                if g1.success_condition is None:
                    g.success_condition = g2.success_condition
                if g1.failure_condition is None:
                    g.failure_condition = g2.failure_condition
                gs.append(g)
    return gs



from __future__ import annotations

from typing import TYPE_CHECKING

import gymnasium as gym
import numpy as np
import torch as th
from stable_baselines3.common.policies import BasePolicy

if TYPE_CHECKING:
    from stable_baselines3.common.type_aliases import PyTorchObs


class RandomPolicy(BasePolicy):
    """
    This class describes a onnx-able policy that returns random actions.
    """

    def _predict(
        self,
        observation: PyTorchObs,
        deterministic: bool = False,
    ) -> th.Tensor:
        if isinstance(observation, dict):
            v = sum(v.sum() for v in observation.values())
            number = next(iter(observation.values())).shape[0]
        else:
            v = observation.sum()
            number = observation.shape[0]
        assert isinstance(self.action_space, gym.spaces.Box)
        high = th.from_numpy(self.action_space.high)
        low = th.from_numpy(self.action_space.low)
        y = th.stack([
            th.rand(self.action_space.shape, device=self.device) *
            (high - low) + low for _ in range(number)
        ])
        y += v * 0
        return y

    def forward(self, x: PyTorchObs, deterministic: bool = False) -> th.Tensor:
        return self._predict(x, deterministic=deterministic)


class SimpleRandomPolicy(BasePolicy):

    def _predict(
        self,
        observation: PyTorchObs,
        deterministic: bool = False,
    ) -> th.Tensor:
        if isinstance(observation, dict):
            number = next(iter(observation.values())).shape[0]
        else:
            number = observation.shape[0]
        actions = np.stack([self.action_space.sample() for _ in range(number)],
                           axis=0)
        return th.as_tensor(actions, device=self.device)

    def forward(self, x: PyTorchObs, deterministic: bool = False) -> th.Tensor:
        return self._predict(x, deterministic=deterministic)



from __future__ import annotations

import abc
import dataclasses as dc
import math
from collections.abc import Mapping
from operator import attrgetter
from typing import TYPE_CHECKING, Any

import numpy as np
from navground import core, sim

from ..register import Registrable
from ..types import Action, Array, Observation

if TYPE_CHECKING:
    import gymnasium as gym

BoxType = type[np.floating[Any]] | type[np.integer[Any]]


@dc.dataclass(repr=False)
class DataclassConfig:
    dtype: str = ''
    """The data type"""

    def __post_init__(self) -> None:
        if self.dtype:
            self._dtype: np.dtype[Any] | None = np.dtype(self.dtype)
            self.dtype = self._dtype.name
        else:
            self._dtype = None

    # Same as dtype accepted in `gym.Box(dtype=...)`
    @property
    def box_type(self) -> BoxType:
        if self._dtype:
            return self._dtype.type
        return core.FloatType

    def __repr__(self) -> str:
        nodef_f_vals = ((f.name, attrgetter(f.name)(self))
                        for f in dc.fields(self)
                        if attrgetter(f.name)(self) != f.default)
        nodef_f_repr = ", ".join(f"{name}={value}"
                                 for name, value in nodef_f_vals)
        return f"{self.__class__.__name__}({nodef_f_repr})"

    def _get_dict(self) -> dict[str, Any]:
        return dc.asdict(self)

    @classmethod
    def _make_from_dict(cls, value: Mapping[str, Any]) -> Any:
        return cls(**value)


@dc.dataclass(repr=False)
class ConfigWithKinematic(DataclassConfig):
    """
    An internal helper class that configures kinematics properties
    """

    max_speed: float = np.inf
    """The maximal speed of the agent"""
    max_angular_speed: float = np.inf
    """The maximal angular speed of the agent"""
    dof: int | None = None
    """The number of degrees of freedom of the agent"""

    def configure_kinematics(self, behavior: core.Behavior) -> None:
        """
        Configure the fields that are not yet set from the behavior's
        kinematics.

        :param      behavior:  The behavior
        """
        kinematics = behavior.kinematics
        if not kinematics:
            return
        if not math.isfinite(self.max_speed):
            self.max_speed = kinematics.max_speed
        if not math.isfinite(self.max_angular_speed):
            self.max_angular_speed = kinematics.max_angular_speed
        if self.dof is None:
            self.dof = kinematics.dof()


class ObservationConfig(abc.ABC, Registrable):
    """
    Configure how to convert the state of a
    a :py:class:`navground.sim.Agent` to observation consumed
    in a :py:class:`gymnasium.Env`.

    Observation may contain data from the agent's behavior ego state (e.g., velocity)
    and :py:attr:`navground.core.Behavior.target` (e.g. relative target direction),
    and data generated by :py:class:`navground.sim.Sensor`.
    """

    @property
    @abc.abstractmethod
    def history(self) -> int:
        """
        The size of observations queue.
        If larger than 1, recent observations will be first
        stacked and then flattened.
        """
        ...

    @abc.abstractmethod
    def configure(self, behavior: core.Behavior | None,
                  sensing_space: gym.spaces.Dict) -> None:
        """
        Complete the configuration for a target behavior is known.

        :param      behavior:  The behavior
        :param      sensing_space:  The sensing space
        """
        ...

    @abc.abstractmethod
    def is_configured(self, warn: bool = False) -> bool:
        """
        Whether is the configuration is complete.

        :returns:   True if complete, False otherwise.
        """
        ...

    # @property
    # @abc.abstractmethod
    # def asdict(self) -> dict[str, Any]:
    #     """
    #     A JSON-able representation of the configuration

    #     :returns:  A JSON-able dict
    #     """
    #     ...

    @abc.abstractmethod
    def get_observation(
            self, behavior: core.Behavior | None,
            buffers: Mapping[str, core.Buffer]) -> Observation:
        ...

    @property
    @abc.abstractmethod
    def space(self) -> gym.Space[Any]:
        ...


@dc.dataclass
class ActionConfig(abc.ABC, Registrable):
    """
    The abstract base time of actions configuration.

    Describes how the agent should convert control commands
    from navground :py:class:`navground.core.Behavior`
    to actions to be applied a :py:class:`gymnasium.Env`
    and, in the opposite direction), actions computed by a ML-policy
    to control commands to be applied in navground.
    """

    @abc.abstractmethod
    def get_cmd_from_action(self, action: Action,
                            behavior: core.Behavior | None,
                            time_step: float) -> core.Twist2:
        """
        Convert action to navground command.

        :param      action:     The action
        :param      behavior:   The behavior that is requesting the command
        :param      time_step:  The time step

        :returns:   A control command.
        """
        ...

    @abc.abstractmethod
    def get_action(self, behavior: core.Behavior, time_step: float) -> Action:
        """
        Gets the action from the command computed by a behavior.

        :param      behavior:   The behavior
        :param      time_step:  The time step

        :returns:   The action.
        """
        ...

    @property
    @abc.abstractmethod
    def space(self) -> gym.Space[Any]:
        """
        The action space

        :returns:   A valid Gymnasium space
        """
        ...

    @abc.abstractmethod
    def configure(self, behavior: core.Behavior) -> None:
        """
        Complete the configuration for a target behavior is known.

        :param      behavior:  The behavior
        """
        ...

    @abc.abstractmethod
    def is_configured(self, warn: bool = False) -> bool:
        """
        Whether is the configuration is complete.

        :returns:   True if complete, False otherwise.
        """
        ...

    # @property
    # @abc.abstractmethod
    # def asdict(self) -> dict[str, Any]:
    #     """
    #     A JSON-able representation of the configuration

    #     :returns:  A JSON-able dict
    #     """
    #     ...


class StateConfig(abc.ABC, Registrable):
    """
    Configure how to get the environment state from the current state of
    a :py:class:`navground.sim.World`
    """

    @abc.abstractmethod
    def get_state(self, world: sim.World) -> Array:
        """
          Computes the global state from the current world state

          :param      world:  The world

          :returns: an array with the global state
        """
        ...

    @abc.abstractmethod
    def get_space(self, world: sim.World) -> gym.spaces.Box:
        """
        Returns the space of the global state

        :param      world:  The world

        :returns:   The space.
        """
        ...



from __future__ import annotations

import dataclasses as dc
from typing import Any

import gymnasium as gym
import numpy as np
from navground import core

from ..config import ModulationActionConfig
from ..types import Action


@dc.dataclass
class DiscreteModulationActionConfig(ModulationActionConfig):

    param: str = ''
    max_index: int = 1

    def __post_init__(self) -> None:
        self.params = {self.param: {'low': 0, 'high': self.max_index, 'discrete': True}}

    @property
    def space(self) -> gym.Space[Any]:
        return gym.spaces.Discrete(n=self.max_index + 1)

    def get_params_from_action(self, action: Action) -> dict[str, int]:
        try:
            value = int(action)
        except Exception:
            value = int(action[0])
        return {self.param: value}

    def get_action(self, behavior: core.Behavior,
                   time_step: float) -> Action:
        return np.array(getattr(behavior, self.param), dtype=int)

    def configure(self, behavior: core.Behavior) -> None:
        pass

    @property
    def asdict(self) -> dict[str, Any]:
        rs = dc.asdict(self)
        rs['type'] = self.__class__.__name__
        return rs



# Patch of https://github.com/HumanCompatibleAI/imitation/blob/master/src/imitation/data/rollout.py

from __future__ import annotations

import dataclasses
import inspect
import logging
import warnings
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, cast, TypeAlias

import gymnasium as gym
import numpy as np
import numpy.typing
from imitation.data import rollout as rollout_without_info
from imitation.data import types
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.policies import BasePolicy
from stable_baselines3.common.utils import check_for_correct_spaces

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv

Array: TypeAlias = np.typing.NDArray[Any]


class TrajectoryAccumulator(rollout_without_info.TrajectoryAccumulator):

    def add_steps_and_auto_finish(
        self,
        acts: Array,
        obs: types.Observation | dict[str, Array],
        rews: np.typing.NDArray[np.floating[Any]],
        dones: np.typing.NDArray[np.bool_],
        infos: list[dict[str, Any]],
    ) -> list[types.TrajectoryWithRew]:
        trajs: list[types.TrajectoryWithRew] = []
        wrapped_obs = types.maybe_wrap_in_dictobs(obs)

        # iterate through environments
        for env_idx in range(len(wrapped_obs)):
            assert env_idx in self.partial_trajectories
            assert list(self.partial_trajectories[env_idx][0].keys()) == [
                "obs"
            ], ("Need to first initialize partial trajectory using "
                "self._traj_accum.add_step({'obs': ob}, key=env_idx)")

        # iterate through steps
        zip_iter = enumerate(
            zip(acts, wrapped_obs, rews, dones, infos, strict=True))
        for env_idx, (act, ob, rew, done, info) in zip_iter:
            if done and "terminal_observation" in info:
                # When dones[i] from VecEnv.step() is True, obs[i] is the first
                # observation following reset() of the ith VecEnv, and
                # infos[i]["terminal_observation"] is the actual final observation.
                real_ob = types.maybe_wrap_in_dictobs(
                    info["terminal_observation"])
            else:
                real_ob = ob

            self.add_step(
                dict(
                    acts=act,
                    rews=rew,
                    # this is not the obs corresponding to `act`, but rather the obs
                    # *after* `act` (see above)
                    obs=real_ob,
                    infos=info,
                ),
                env_idx,
            )
            if done:
                # finish env_idx-th trajectory
                new_traj = self.finish_trajectory(env_idx, terminal=True)
                trajs.append(new_traj)
                # When done[i] from VecEnv.step() is True, obs[i] is the first
                # observation following reset() of the ith VecEnv.
                self.add_step(dict(obs=ob), env_idx)
        return trajs


# A PolicyCallableWithInfo is a function that takes an array of observations, an optional
# array of states, an optional array of episode starts, *and an optional list of infos*,
# and returns an array of corresponding actions.

PolicyCallableWithInfo: TypeAlias = Callable[
    [
        Array | dict[str, Array],  # observations
        tuple[Array, ...] | None,  # states
        np.typing.NDArray[np.bool_] | None,  # episode_starts
        list[dict[str, Array]] | None  # (Jerome) added info
    ],
    tuple[Array, tuple[Array, ...] | None]  # actions, states
]

AnyPolicy: TypeAlias = rollout_without_info.AnyPolicy | PolicyCallableWithInfo


def policy_to_callable(
    policy: AnyPolicy,
    venv: VecEnv,
    deterministic_policy: bool = False,
) -> PolicyCallableWithInfo:
    """Converts any policy-like object into a function from observations to actions."""
    get_actions: PolicyCallableWithInfo
    if policy is None:

        def get_actions(
            observations: Array | dict[str, Array],
            states: tuple[Array, ...] | None,
            episode_starts: np.typing.NDArray[np.bool_] | None,
            infos: list[dict[str, Array]] | None = None
        ) -> tuple[Array, tuple[Array, ...] | None]:
            acts = [
                venv.action_space.sample() for _ in range(len(observations))
            ]
            return np.stack(acts, axis=0), None

    elif isinstance(policy, (BaseAlgorithm, BasePolicy)):
        # There's an important subtlety here: BaseAlgorithm and BasePolicy
        # are themselves Callable (which we check next). But in their case,
        # we want to use the .predict() method, rather than __call__()
        # (which would call .forward()). So this elif clause must come first!

        def get_actions(
            observations: Array | dict[str, Array],
            states: tuple[Array, ...] | None,
            episode_starts: np.typing.NDArray[np.bool_] | None,
            infos: list[dict[str, Array]] | None = None
        ) -> tuple[Array, tuple[Array, ...] | None]:
            assert isinstance(policy, (BaseAlgorithm, BasePolicy))
            # pytype doesn't seem to understand that policy is a BaseAlgorithm
            # or BasePolicy here, rather than a Callable
            (acts, states) = policy.predict(  # pytype: disable=attribute-error
                observations,
                state=states,
                episode_start=episode_starts,
                deterministic=deterministic_policy,
            )
            return acts, states

    elif callable(policy):
        # When a policy callable is passed, by default we will use it directly.
        # We are not able to change the determinism of the policy when it is a
        # callable that only takes in the states.
        if deterministic_policy:
            # raise ValueError(
            warnings.warn(
                "Cannot set deterministic_policy=True when policy is a callable, "
                "since deterministic_policy argument is ignored.",
                stacklevel=1)

        sig = inspect.signature(policy)
        if len(sig.parameters) == 3:

            def get_actions(
                observations: Array | dict[str, Array],
                states: tuple[Array, ...] | None,
                episode_starts: np.typing.NDArray[np.bool_] | None,
                infos: list[dict[str, Array]] | None = None
            ) -> tuple[Array, tuple[Array, ...] | None]:
                return cast("rollout_without_info.PolicyCallable",
                            policy)(observations, states, episode_starts)
        else:
            get_actions = cast("PolicyCallableWithInfo", policy)
    else:
        raise TypeError(
            "Policy must be None, a stable-baselines policy or algorithm, "
            f"or a Callable, got {type(policy)} instead", )

    if isinstance(policy, BaseAlgorithm):
        # Check that the observation and action spaces of policy and environment match
        try:
            check_for_correct_spaces(
                venv,
                policy.observation_space,
                policy.action_space,
            )
        except ValueError as e:
            # Check for a particularly common mistake when using image environments.
            venv_obs_shape = venv.observation_space.shape
            assert policy.observation_space is not None
            policy_obs_shape = policy.observation_space.shape
            assert venv_obs_shape is not None
            assert policy_obs_shape is not None
            if len(venv_obs_shape) != 3 or len(policy_obs_shape) != 3:
                raise e
            venv_obs_rearranged = (
                venv_obs_shape[2],
                venv_obs_shape[0],
                venv_obs_shape[1],
            )
            if venv_obs_rearranged != policy_obs_shape:
                raise e
            raise ValueError(
                "Policy and environment observation shape mismatch. "
                "This is likely caused by "
                "https://github.com/HumanCompatibleAI/imitation/issues/599. "
                "If encountering this from rollout.rollout, try calling:\n"
                "rollout.rollout(expert, expert.get_env(), ...) instead of\n"
                "rollout.rollout(expert, env, ...)\n\n"
                f"Policy observation shape: {policy_obs_shape} \n"
                f"Environment observation shape: {venv_obs_shape}") from e

    return get_actions


def generate_trajectories(
    policy: AnyPolicy,
    venv: VecEnv,
    sample_until: rollout_without_info.GenTrajTerminationFn,
    rng: np.random.Generator,
    *,
    deterministic_policy: bool = False,
) -> Sequence[types.TrajectoryWithRew]:
    """
    Patches the orginal version in ``imitation``
    to add support for :py:type:`PolicyCallableWithInfo`

    :param      policy:                The policy
    :param      venv:                  The venv
    :param      sample_until:          The sample until criteria
    :param      rng:                   The random number generator
    :param      deterministic_policy:  Whether the policy is deterministic

    :returns:   the trajectories
    """
    get_actions = policy_to_callable(policy, venv, deterministic_policy)

    # Collect rollout tuples.
    trajectories = []
    # accumulator for incomplete trajectories
    trajectories_accum = TrajectoryAccumulator(
    )  # type: ignore[no-untyped-call]
    obs = venv.reset()
    infos = venv.reset_infos
    assert isinstance(
        obs, (np.ndarray, dict)), "Tuple observations are not supported."
    wrapped_obs = types.maybe_wrap_in_dictobs(obs)

    # we use dictobs to iterate over the envs in a vecenv
    for env_idx, ob in enumerate(wrapped_obs):
        # Seed with first obs only. Inside loop, we'll only add second obs from
        # each (s,a,r,s') tuple, under the same "obs" key again. That way we still
        # get all observations, but they're not duplicated into "next obs" and
        # "previous obs" (this matters for, e.g., Atari, where observations are
        # really big).
        trajectories_accum.add_step(dict(obs=ob), env_idx)

    # Now, we sample until `sample_until(trajectories)` is true.
    # If we just stopped then this would introduce a bias towards shorter episodes,
    # since longer episodes are more likely to still be active, i.e. in the process
    # of being sampled from. To avoid this, we continue sampling until all epsiodes
    # are complete.
    #
    # To start with, all environments are active.
    active = np.ones(venv.num_envs, dtype=bool)
    state = None
    dones = np.zeros(venv.num_envs, dtype=bool)
    while np.any(active):
        if any(dones):
            infos = list(infos)
            for i, v in enumerate(dones):
                if v:
                    infos[i] = venv.reset_infos[i]
        # policy gets unwrapped observations (eg as dict, not dictobs)
        acts, state = get_actions(obs, state, dones, infos)
        obs, rews, dones, infos = venv.step(acts)
        dones = dones.astype(bool)
        assert isinstance(
            obs,
            (np.ndarray, dict),
        ), "Tuple observations are not supported."
        wrapped_obs = types.maybe_wrap_in_dictobs(obs)

        # If an environment is inactive, i.e. the episode completed for that
        # environment after `sample_until(trajectories)` was true, then we do
        # *not* want to add any subsequent trajectories from it. We avoid this
        # by just making it never done.
        dones &= active

        new_trajs = trajectories_accum.add_steps_and_auto_finish(
            acts,
            wrapped_obs,
            rews,
            dones,
            infos,
        )
        trajectories.extend(new_trajs)

        if sample_until(trajectories):
            # Termination condition has been reached. Mark as inactive any
            # environments where a trajectory was completed this timestep.
            active &= ~dones

    # Note that we just drop partial trajectories. This is not ideal for some
    # algos; e.g. BC can probably benefit from partial trajectories, too.

    # Each trajectory is sampled i.i.d.; however, shorter episodes are added to
    # `trajectories` sooner. Shuffle to avoid bias in order. This is important
    # when callees end up truncating the number of trajectories or transitions.
    # It is also cheap, since we're just shuffling pointers.
    rng.shuffle(trajectories)  # type: ignore[arg-type]

    # Sanity checks.
    for trajectory in trajectories:
        n_steps = len(trajectory.acts)
        # extra 1 for the end
        if isinstance(venv.observation_space, gym.spaces.Dict):
            exp_obs = {}
            for k, v in venv.observation_space.items():
                assert v.shape is not None
                exp_obs[k] = (n_steps + 1, ) + v.shape
        else:
            obs_space_shape = venv.observation_space.shape
            assert obs_space_shape is not None
            exp_obs = (n_steps +
                       1, ) + obs_space_shape  # type: ignore[assignment]
        real_obs = trajectory.obs.shape
        assert real_obs == exp_obs, f"expected shape {exp_obs}, got {real_obs}"
        assert venv.action_space.shape is not None
        exp_act = (n_steps, ) + venv.action_space.shape
        real_act = trajectory.acts.shape
        assert real_act == exp_act, f"expected shape {exp_act}, got {real_act}"
        exp_rew = (n_steps, )
        real_rew = trajectory.rews.shape
        assert real_rew == exp_rew, f"expected shape {exp_rew}, got {real_rew}"

    return trajectories


def generate_transitions(
    policy: AnyPolicy,
    venv: VecEnv,
    n_timesteps: int,
    rng: np.random.Generator,
    *,
    truncate: bool = True,
    **kwargs: Any,
) -> types.TransitionsWithRew:
    traj = generate_trajectories(
        policy,
        venv,
        sample_until=rollout_without_info.make_min_timesteps(n_timesteps),
        rng=rng,
        **kwargs,
    )
    transitions = rollout_without_info.flatten_trajectories_with_rew(traj)
    if truncate and n_timesteps is not None:
        as_dict = types.dataclass_quick_asdict(transitions)
        truncated = {k: arr[:n_timesteps] for k, arr in as_dict.items()}
        transitions = types.TransitionsWithRew(**truncated)
    return transitions


def rollout(
    policy: AnyPolicy,
    venv: VecEnv,
    sample_until: rollout_without_info.GenTrajTerminationFn,
    rng: np.random.Generator,
    *,
    unwrap: bool = True,
    exclude_infos: bool = True,
    verbose: bool = True,
    **kwargs: Any,
) -> Sequence[types.TrajectoryWithRew]:
    trajs = generate_trajectories(
        policy,
        venv,
        sample_until,
        rng=rng,
        **kwargs,
    )
    if unwrap:
        trajs = [rollout_without_info.unwrap_traj(traj) for traj in trajs]
    if exclude_infos:
        trajs = [dataclasses.replace(traj, infos=None) for traj in trajs]
    if verbose:
        stats = rollout_without_info.rollout_stats(trajs)
        logging.info(f"Rollout stats: {stats}")
    return trajs



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

import gymnasium as gym
import torch as th
from navground.core import FloatType
from stable_baselines3.common.distributions import (
    SquashedDiagGaussianDistribution, StateDependentNoiseDistribution)
from stable_baselines3.common.torch_layers import (CombinedExtractor,
                                                   FlattenExtractor,
                                                   create_mlp)
from stable_baselines3.sac.policies import (LOG_STD_MAX, LOG_STD_MIN, Actor,
                                            BasePolicy, SACPolicy)
from torch import nn

if TYPE_CHECKING:

    from stable_baselines3.common.torch_layers import BaseFeaturesExtractor
    from stable_baselines3.common.type_aliases import (PyTorchObs, Schedule,
                                                       TensorDict)


class StackCombinedExtractor(CombinedExtractor):

    def forward(self, observations: TensorDict) -> th.Tensor:
        encoded_tensor_list = []
        for key, extractor in self.extractors.items():
            encoded_tensor_list.append(extractor(observations[key]))
        return th.stack(encoded_tensor_list, dim=-1).flatten(start_dim=1)


class ActorWithComm(Actor):

    def __init__(self,
                 observation_space: gym.Space,
                 action_space: gym.spaces.Box,
                 net_arch: list[int],
                 features_extractor: nn.Module,
                 features_dim: int,
                 activation_fn: type[nn.Module] = nn.ReLU,
                 use_sde: bool = False,
                 log_std_init: float = -3,
                 full_std: bool = True,
                 use_expln: bool = False,
                 clip_mean: float = 2.0,
                 normalize_images: bool = True,
                 comm_space: gym.spaces.Box = gym.spaces.Box(low=-1,
                                                             high=1,
                                                             dtype=FloatType),
                 comm_net_arch: list[int] = [32, 32]):
        self.comm_space = comm_space
        self.comm_size = gym.spaces.utils.flatdim(comm_space)
        self.comm_net_arch = net_arch[:]
        super().__init__(observation_space,
                         action_space,
                         net_arch,
                         features_extractor,
                         features_dim=features_dim,
                         activation_fn=activation_fn,
                         use_sde=use_sde,
                         log_std_init=log_std_init,
                         full_std=full_std,
                         use_expln=use_expln,
                         clip_mean=clip_mean,
                         normalize_images=normalize_images)
        self._has_dict_space = isinstance(observation_space, gym.spaces.Dict)
        self.number = self.action_space.shape[0]
        self.single_agent_features_dim = features_dim // self.number
        features_dim = self.single_agent_features_dim + self.comm_size
        latent_pi_net = create_mlp(features_dim, -1, net_arch, activation_fn)
        self.latent_pi = nn.Sequential(*latent_pi_net)

        tx_net = create_mlp(self.single_agent_features_dim, -1,
                            [*self.comm_net_arch, self.comm_size], nn.Sigmoid)
        self.tx_net = nn.Sequential(*tx_net)
        action_dim = self.action_space.shape[1]
        last_layer_dim = net_arch[-1] if len(net_arch) > 0 else features_dim
        if self.use_sde:
            self.action_dist = StateDependentNoiseDistribution(
                action_dim,
                full_std=full_std,
                use_expln=use_expln,
                learn_features=True,
                squash_output=True)
            self.mu, self.log_std = self.action_dist.proba_distribution_net(
                latent_dim=last_layer_dim,
                latent_sde_dim=last_layer_dim,
                log_std_init=log_std_init)
            if clip_mean > 0.0:
                self.mu = nn.Sequential(
                    self.mu, nn.Hardtanh(min_val=-clip_mean,
                                         max_val=clip_mean))
        else:
            self.action_dist = SquashedDiagGaussianDistribution(
                action_dim)  # type: ignore[assignment]
            self.mu = nn.Linear(last_layer_dim, action_dim)
            self.log_std = nn.Linear(last_layer_dim,
                                     action_dim)  # type: ignore[assignment]

    def get_action_dist_params_after_extraction(
        self, features: th.Tensor
    ) -> tuple[th.Tensor, th.Tensor, dict[str, th.Tensor]]:
        latent_pi = self.latent_pi(features)
        mean_actions = self.mu(latent_pi)
        if self.use_sde:
            return mean_actions, self.log_std, dict(latent_sde=latent_pi)
        log_std = self.log_std(latent_pi)  # type: ignore[operator]
        log_std = th.clamp(log_std, LOG_STD_MIN, LOG_STD_MAX)
        return mean_actions, log_std, {}

    def scale_comm(self, values: th.Tensor) -> th.Tensor:
        # TODO(Jerome): generalize to non-homogeneous comm spaces
        low, high = self.comm_space.low[0], self.comm_space.high[0]
        return low + values * (high - low)

    def get_action_dist_params(
            self, obs: PyTorchObs
    ) -> tuple[th.Tensor, th.Tensor, dict[str, th.Tensor]]:
        features = self.extract_features(obs, self.features_extractor)
        features = features.reshape(
            (-1, self.number, self.single_agent_features_dim))
        tx = self.tx_net(features)
        tx = self.scale_comm(tx)
        self._last_tx = tx.detach().numpy()
        rx = []
        for i in range(self.number - 1):
            rx.append(th.roll(tx, i + 1, dims=1))
        # self._last_rxs = [r.detach().numpy() for r in rx]

        features = th.concat([features, *rx], dim=-1)
        features = features.reshape((-1, features.shape[-1]))
        m, le, info = self.get_action_dist_params_after_extraction(features)
        shape = cast('gym.spaces.Box', self.action_space).shape
        s = shape[0] * shape[1]
        return m.reshape((-1, s)), le.reshape((-1, s)), info


class SACPolicyWithComm(SACPolicy):
    """
    This class describes a SB3 Sac policy that works
    on a stacked multi-agent environments.
    (:py:class:`navground.learning.parallel_env.JointEnv`)

    The actor is composed of two modules: CommNet and ActionNet.
    CommNet takes stacked (i.e. batched) single agent observations and
    computes a message (for each agent). ActionNet takes stacked (i.e. batched)
    single agent observations and all other agents messages and computes actions.

    Communication never exit the policy during training.
    During inference, we can evaluate the two sub-networks separately
    and explicitly share the messages, using :py:class:`DistributedCommPolicy`.

    The class supports composed and simple observations, selecting the corresponding
    features extractor at initialization.

    User can configure the CommNet by including fields

    - comm_space: gym.space.Box
    - comm_net_arch: list[int]

    in ``policy_kwargs``.
    """

    def __init__(
        self,
        observation_space: gym.Space,
        action_space: gym.spaces.Box,
        lr_schedule: Schedule,
        net_arch: list[int] | dict[str, list[int]] | None = None,
        activation_fn: type[nn.Module] = nn.ReLU,
        use_sde: bool = False,
        log_std_init: float = -3,
        use_expln: bool = False,
        clip_mean: float = 2.0,
        features_extractor_class: type[BaseFeaturesExtractor] | None = None,
        features_extractor_kwargs: dict[str, Any] | None = None,
        normalize_images: bool = True,
        optimizer_class: type[th.optim.Optimizer] = th.optim.Adam,
        optimizer_kwargs: dict[str, Any] | None = None,
        n_critics: int = 2,
        share_features_extractor: bool = False,
        comm_space: gym.spaces.Box = gym.spaces.Box(low=-1,
                                                    high=1,
                                                    dtype=FloatType),
        comm_net_arch: list[int] = [32, 32],
    ):
        """
        Constructs a new instance.

        Same parameters of the super class :py:class:`SACPolicy` constructor,
        with the addition of ``actor_specs``.

        :param      comm_space:    The comm space.
        :param      comm_net_arch: The comm network MLP specs
        """
        self._comm_space = comm_space
        self._comm_net_arch = comm_net_arch
        if features_extractor_class is None:
            if isinstance(observation_space, gym.spaces.Dict):
                features_extractor_class = StackCombinedExtractor
            else:
                features_extractor_class = FlattenExtractor
        SACPolicy.__init__(
            self,
            observation_space,
            action_space,
            lr_schedule,
            net_arch,
            activation_fn,
            use_sde,
            log_std_init,
            use_expln,
            clip_mean,
            features_extractor_class,
            features_extractor_kwargs,
            normalize_images,
            optimizer_class,
            optimizer_kwargs,
            n_critics,
            share_features_extractor,
        )

    def make_actor(
            self,
            features_extractor: BaseFeaturesExtractor | None = None) -> Actor:
        actor_kwargs = self._update_features_extractor(self.actor_kwargs,
                                                       features_extractor)
        return ActorWithComm(
            comm_space=self._comm_space,
            comm_net_arch=self._comm_net_arch,
            **actor_kwargs,
        ).to(self.device)


class DistributedCommPolicy(BasePolicy):
    """
    This class converts a :py:class:`SACPolicyWithComm` centralized policy
    in a distributed policy, evaluating the two sub-networks
    to compute action and outgoing message, which are returned concatenated.
    """

    def __init__(self, observation_space: gym.spaces.Dict | gym.spaces.Box,
                 action_space: gym.spaces.Box, policy: SACPolicyWithComm,
                 **kwargs: Any):
        """
        Constructs a new instance.

        :param      observation_space:  The observation space
        :param      action_space:       The action space
        :param      policy:             The centralized policy
        :param      kwargs:             The keywords arguments passed
                                        to the super class
        """
        super().__init__(observation_space, action_space, **kwargs)
        assert isinstance(policy.actor, ActorWithComm)
        self._actor: ActorWithComm = policy.actor
        if isinstance(observation_space, gym.spaces.Dict):
            extra_keys = set(observation_space.spaces) - set(
                cast('gym.spaces.Dict', self._actor.observation_space).spaces)
            assert len(extra_keys) == 1
            self._has_dict_space = True
            self.comm_key = extra_keys.pop()
            self.comm_extractor = nn.Flatten()

    def forward(self,
                observation: PyTorchObs,
                deterministic: bool = False) -> th.Tensor:
        return self._predict(observation, deterministic=deterministic)

    def _predict(self,
                 observation: PyTorchObs,
                 deterministic: bool = False) -> th.Tensor:
        if self._has_dict_space:
            comm_observation = cast('dict[str, th.Tensor]',
                                    observation).pop(self.comm_key)
            features_without_comm = self._actor.extract_features(
                observation, self._actor.features_extractor)
            features_comm = self.comm_extractor(comm_observation).reshape(
                *features_without_comm.shape[:-1], -1)
            features = th.cat([features_without_comm, features_comm], dim=-1)
        else:
            features = cast('th.Tensor', observation)
            features_without_comm = features[
                ..., :self._actor.single_agent_features_dim]
        tx = self._actor.tx_net(features_without_comm)
        mean_actions, log_std, kwargs = self._actor.get_action_dist_params_after_extraction(
            features)
        act = self._actor.action_dist.actions_from_params(
            mean_actions, log_std, deterministic=deterministic, **kwargs)
        space = cast('gym.spaces.Box', self.action_space)
        low, high = space.low[0], space.high[0]
        act = low + (0.5 * (act + 1.0) * (high - low))
        tx = self._actor.scale_comm(tx)
        return th.concat([act, tx], dim=-1)

    def set_training_mode(self, mode: bool) -> None:
        pass



from __future__ import annotations

import dataclasses as dc
from collections.abc import Iterable
from typing import Any, cast

import gymnasium as gym
import numpy as np
from navground import core

from ..types import Action
from .control_action import ControlActionConfig


def to_discrete(values: Iterable[float], threshold: float) -> int:
    r = 0
    for i, value in enumerate(values):
        x = 1
        if value > threshold:
            x = 2
        if value < threshold:
            x = 0
        r += (3**i) * x
    return r


def to_continuous(value: int, number: int) -> list[float]:
    rs = []
    for _ in range(number):
        rs.append((value % 3) / 2)
        value = value // 3
    return rs


# TODO(Jerome): handle the case when the continuous low is 0,
# where we need only 2 values instead of 3.

@dc.dataclass
class DiscreteControlActionConfig(ControlActionConfig,
                                  register_name="DiscreteControl"):
    """
    Configuration of the conversion between *discrete* control actions
    and control commands. Actions are either command accelerations
    or command velocities, depending on :py:attr:`ControlActionConfig.use_acceleration_action`.

    Discrete actions are first converted to continuous action and then,
    by the super class :py:class:`ControlActionConfig`, to control commands.

    The action space :py:class:`gymnasium.spaces.Discrete` is discrete.
    Values are encoded using a ternary numeral system,
    where for every dimension:

    .. list-table:: Title
       :header-rows: 1

       * - Discrete
         - Continuous
       * - 0
         - ``space.low``
       * - 1
         - 0
       * - 2
         - ``space.high``

    :param dtype: The data type

    :param dof: The number of degrees of freedom of the agent

    :param max_speed: The upper bound of the speed.

    :param max_angular_speed: The upper bound of the angular speed.

    :param max_acceleration: The upper bound of the acceleration.

    :param max_angular_acceleration: The upper bound of the angular acceleration.

    :param use_acceleration_action: Whether actions are accelerations.
                                    If not set, actions are velocities.

    :param use_wheels: Whether action uses wheel speeds/acceleration
                       instead of body speeds/acceleration.
                       Only effective if the b behavior has a wheeled kinematics.

    :param fix_orientation: Whether to force the agent not to control orientation,
                            i.e., to not include the angular command in actions.

    :param has_wheels: Whether the agent as wheels.
                       If None, it will defer to the agent kinematics.

    """

    def configure(self, behavior: core.Behavior) -> None:
        super().configure(behavior)
        self._cont_space = cast('gym.spaces.Box', super().space)

    def get_cmd_from_action(self, action: Action,
                            behavior: core.Behavior | None,
                            time_step: float) -> core.Twist2:
        act = np.array(to_continuous(int(action), self.action_size),
                       dtype=self._cont_space.dtype)
        act = self._cont_space.low + act * (self._cont_space.high -
                                            self._cont_space.low)
        return super().get_cmd_from_action(act, behavior, time_step)

    def get_action(self, behavior: core.Behavior, time_step: float) -> Action:
        act = ControlActionConfig.get_action(self, behavior, time_step)
        th = (self._cont_space.high + self._cont_space.low) / 2
        return np.array(to_discrete(act, threshold=th),
                        dtype=np.int64).flatten()

    @property
    def space(self) -> gym.Space[Any]:
        """
        The action space.

        - 0: backwards
        - 1: do-nothing
        - 2: forwards
        """
        return gym.spaces.Discrete(3**self.action_size)



from __future__ import annotations

import warnings
from collections.abc import Mapping
from typing import Any, TypeAlias

import gymnasium as gym
from gymnasium.envs.registration import register
from navground import sim

from ..config import (ActionConfig, ControlActionConfig,
                      DefaultObservationConfig, GroupConfig, ObservationConfig)
from ..indices import Indices
from ..internal.base_env import NavgroundBaseEnv
from ..policies.info_predictor import InfoPolicy
from ..rewards import NullReward
from ..types import (Action, Bounds, Observation, Reward, SensorLike,
                     SensorSequenceLike, TerminationCondition)

BaseEnv: TypeAlias = gym.Env[Observation, Action]


class NavgroundEnv(NavgroundBaseEnv, BaseEnv):
    """
    This class describes an environment that uses
    a :py:class:`navground.sim.Scenario` to
    generate and then simulate a :py:class:`navground.sim.World`.

    Actions and observations relates to a single selected
    individual :py:class:`navground.sim.Agent`.

    The behavior is registered under the id ``"navground"``:

    >>> import gymnasium as gym
    >>> import navground.learning.env
    >>> from navground import sim
    >>>
    >>> scenario = sim.load_scenario(...)
    >>> env = gym.make("navground", scenario=scenario)

    :param scenario: The scenario to initialize all simulated worlds.
                     If a :py:class:`str`, it will be interpreted as the YAML
                     representation of a scenario.
                     If a :py:class:`dict`, it will be dumped to YAML and
                     then treated as a :py:class:`str`.

    :param agent_index: The world index of the selected agent,
                        must be smaller than the number of agents.

    :param sensor: An optional sensor that will be added to ``sensors``.

    :param sensors: A sequence of sensor to generate observations for the agents
                    or its YAML representation. If
                    Items of class :py:class:`str` will be interpreted as the YAML
                    representation of a sensor.
                    Items of class :py:class:`dict` will be dumped to YAML and
                    then treated as a :py:class:`str`.
                    If empty, it will use the agents' own sensors.

    :param action: The configuration of the action space to use.

    :param observation: The configuration of the observation space to use.

    :param reward: The reward function to use. If none, it will default to constant zeros.

    :param time_step: The simulation time step applied at every :py:meth:`step`.

    :param max_duration: If positive, it will signal a truncation after this simulated time.

    :param terminate_if_idle: Whether to terminate when an agent is idle

    :param truncate_outside_bounds: Whether to truncate when an agent exit the bounds

    :param bounds: The area to render and a fence for truncating processes when agents exit it.

    :param render_mode: The render mode.
                        If `"human"`, it renders a simulation in real time via
                        websockets (see :py:class:`navground.sim.ui.WebUI`).
                        If `"rgb_array"`, it uses
                        :py:func:`navground.sim.ui.render.image_for_world`
                        to render the world on demand.

    :param render_kwargs: Arguments passed to :py:func:`navground.sim.ui.render.image_for_world`

    :param realtime_factor: a realtime factor for `render_mode="human"`: larger values
                            speed up the simulation.

    :param stuck_timeout: The time to wait before considering an agent stuck and terminate it.

    :param color: An optional color of the agent (only used for displaying)

    :param tag: An optional tag to be added to the agent (only used as metadata)

    :param terminate_on_success: Whether to terminate the episode on success.

    :param terminate_on_failure: Whether to terminate the episode on failure.

    :param success_condition: An optional success criteria

    :param failure_condition: An optional failure criteria

    :param include_action: Whether to include field "navground_action" in the info

    :param include_success: Whether to include field "is_success" in the info

    :param init_success: The default value of success (valid until a termination condition is met)

    :param intermediate_success: Whether to include "is_success" in the info at intermediate
                                 steps (vs only at termination).
    """

    metadata = {"render_modes": ["human", "rgb_array"], 'render_fps': 30}
    MAX_SEED = 2**31 - 1

    def __init__(self,
                 scenario: sim.Scenario | str | dict[str, Any] | None = None,
                 agent_index: int = 0,
                 sensor: SensorLike | None = None,
                 sensors: SensorSequenceLike = tuple(),
                 action: ActionConfig = ControlActionConfig(),
                 observation: ObservationConfig = DefaultObservationConfig(),
                 reward: Reward | None = None,
                 time_step: float = 0.1,
                 max_duration: float = -1.0,
                 terminate_if_idle: bool = True,
                 bounds: Bounds | None = None,
                 truncate_outside_bounds: bool = False,
                 render_mode: str | None = None,
                 render_kwargs: Mapping[str, Any] = {},
                 realtime_factor: float = 1.0,
                 stuck_timeout: float = -1,
                 color: str = '',
                 tag: str = '',
                 terminate_on_success: bool = True,
                 terminate_on_failure: bool = True,
                 success_condition: TerminationCondition | None = None,
                 failure_condition: TerminationCondition | None = None,
                 include_action: bool = True,
                 include_success: bool = True,
                 init_success: bool | None = None,
                 intermediate_success: bool = False) -> None:
        if reward is None:
            reward = NullReward()
        self.agent_index = agent_index
        group = GroupConfig(indices=Indices({agent_index}),
                            sensor=sensor,
                            sensors=sensors,
                            reward=reward,
                            action=action,
                            observation=observation,
                            color=color,
                            tag=tag,
                            terminate_on_success=terminate_on_success,
                            terminate_on_failure=terminate_on_failure,
                            success_condition=success_condition,
                            failure_condition=failure_condition)
        NavgroundBaseEnv.__init__(
            self,
            groups=[group],
            scenario=scenario,
            time_step=time_step,
            max_duration=max_duration,
            terminate_if_idle=terminate_if_idle,
            bounds=bounds,
            truncate_outside_bounds=truncate_outside_bounds,
            render_mode=render_mode,
            render_kwargs=render_kwargs,
            realtime_factor=realtime_factor,
            stuck_timeout=stuck_timeout,
            include_action=include_action,
            include_success=include_success,
            init_success=init_success,
            intermediate_success=intermediate_success)

    def _init(self) -> None:
        super()._init()
        if self.scenario is not None:
            agent = self._possible_agents[self.agent_index]
            with warnings.catch_warnings(record=True) as ws:
                warnings.simplefilter("always")
                configured = agent.is_configured(warn=True)
            if configured:
                self.observation_space = self._observation_space[
                    self.agent_index]
                self.action_space = self._action_space[self.agent_index]
            else:
                msgs = ', '.join([str(w.message) for w in ws])
                warnings.warn(
                    f"Configuration of agent at {self.agent_index} "
                    "is not complete. Check that the scenario spawns "
                    "the agent and that the action and observation "
                    f"configs have the required information: {msgs}",
                    stacklevel=0)
                self.observation_space = gym.spaces.Box(0, 1)
                self.action_space = gym.spaces.Box(0, 1)
            self.reward = agent.reward
        else:
            self.observation_space = gym.spaces.Box(0, 1)
            self.action_space = gym.spaces.Box(0, 1)
            self.reward = None

    def _init_spec(
            self,
            scenario: sim.Scenario | str | dict[str, Any] | None) -> None:
        super()._init_spec(scenario)
        self._spec.pop('groups')
        group = next(iter(self.groups_config))
        self._spec['agent_index'] = self.agent_index
        self._spec['reward'] = group.reward
        self._spec['action'] = group.action
        self._spec['observation'] = group.observation
        self._spec['sensors'] = group.sensors
        self._spec['sensor'] = None

    @property
    def policy(self) -> InfoPolicy:
        """
        A policy that returns the action computed by the navground agent.

        :returns:   The policy.
        """
        return self.get_policy(self.agent_index)

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None
    ) -> tuple[Observation, dict[str, Any]]:
        """
        Conforms to :py:meth:`gymnasium.Env.reset`.

        It samples a new world from a scenario, runs one dry simulation step
        using :py:meth:`navground.sim.World.update_dry`.
        Then, it converts the agent's state to observations.
        and the command it would actuate to an action, which it includes
        in the ``info`` dictionary at key `"navground_action"``.
        """
        gym.Env.reset(self, seed=seed, options=options)
        NavgroundBaseEnv._reset(self, seed, options)
        obs = self.get_observations()
        infos = self.get_infos()
        return obs[self.agent_index], infos[self.agent_index]

    def step(
        self, action: Action
    ) -> tuple[Observation, float, bool, bool, dict[str, Action]]:
        """
        Conforms to :py:meth:`gymnasium.Env.step`.

        It converts the action to a command that the navground agent actuates.
        Then, it updates the world for one step, calling
        :py:meth:`navground.sim.World.update`.
        Finally, it converts the agent's state to observations,
        the command it would actuate to an action, which it includes
        in the ``info`` dictionary at key `"navground_action"`, and computes a reward.

        Termination is set when the agent completes the task, exits the boundary, or gets stuck.
        Truncation is set when the maximal duration has passed.
        """
        obs, reward, terminated, truncated, infos = NavgroundBaseEnv._step(
            self, {self.agent_index: action})
        return (obs[self.agent_index], reward[self.agent_index],
                terminated[self.agent_index], truncated[self.agent_index],
                infos[self.agent_index])


register(
    id="navground",
    entry_point="navground.learning.env:NavgroundEnv",
)



from __future__ import annotations

from typing import Any, cast

import gymnasium as gym
import numpy as np

from ..types import Array


def get_number_of_batches_in_array(value: Array, space: gym.spaces.Box) -> int:
    if len(value.shape) == len(space.shape):
        return 0
    if len(value.shape) == len(space.shape) + 1:
        return cast("int", value.shape[0])
    raise ValueError(f"Invalid shape {value.shape} for space {space}")


def get_number_of_batches_in_dict(value: dict[str, Array],
                                  space: gym.spaces.Dict) -> int:
    keys = set(space)
    batches = [
        get_number_of_batches_in_array(v, cast("gym.spaces.Box", space[k]))
        for k, v in value.items()
        if k in keys and isinstance(space[k], gym.spaces.Box)
    ]
    if not batches or len(set(batches)) > 1:
        raise ValueError(f"Invalid value {value} for space {space}")
    return batches[0]


def get_number_of_batches(value: Array | dict[str, Array],
                          space: gym.Space[Any]) -> int:
    if isinstance(space, gym.spaces.Dict):
        if isinstance(value, dict):
            return get_number_of_batches_in_dict(value, space)
        raise TypeError(f"Wrong type {type(value)} for Dict space")
    if isinstance(space, gym.spaces.Box):
        if isinstance(value, np.ndarray):
            return get_number_of_batches_in_array(value, space)
        raise TypeError(f"Wrong type {type(value)} for Box space")
    raise TypeError(f"Space of type {type(space)} not supported")



from __future__ import annotations

import functools
from collections.abc import Callable
from typing import cast

import numpy as np
from navground import sim

from ..indices import Indices, IndicesLike

Condition = Callable[[sim.Agent, sim.World], bool]


class SuccessProbe(sim.RecordProbe):
    """
    A probe that records successes statuses to a single dataset
    of shape ``(#agents, )``, where

    - -1 = unknown
    - 0 = failure
    - 1 = success

    :param      ds:         The dataset
    :param      indices:    The indices of the agents to record
    :param      is_success: Success condition
    :param      is_success: Failure condition
    """

    dtype = np.int8

    def __init__(self,
                 ds: sim.Dataset,
                 indices: IndicesLike = Indices.all(),
                 is_success: Condition | None = None,
                 is_failure: Condition | None = None):
        super().__init__(ds)
        self._indices = Indices(indices)
        self._is_success = is_success
        self._is_failure = is_failure

    def update(self, run: sim.ExperimentalRun) -> None:
        if self._is_success or self._is_failure:
            for agent in self._indices.sub_sequence(run.world.agents):
                if not hasattr(agent, '_success'):
                    if self._is_success and self._is_success(agent, run.world):
                        agent._success = True  # type: ignore[attr-defined]
                    if self._is_failure and self._is_failure(agent, run.world):
                        agent._success = False  # type: ignore[attr-defined]

    def finalize(self, run: sim.ExperimentalRun) -> None:
        for agent in self._indices.sub_sequence(run.world.agents):
            self.data.push(getattr(agent, '_success', -1))

    def get_shape(self, world: sim.World) -> list[int]:
        return [len(self._indices.sub_sequence(world.agents))]

    @classmethod
    def with_indices(
            cls,
            indices: IndicesLike) -> Callable[[sim.Dataset], SuccessProbe]:
        """
        Creates a probe factory to record success statuses.

        :param      indices:  The indices of the agents to record

        :returns:   A callable that can be added to runs or experiments using
            :py:meth:`navground.sim.ExperimentalRun.add_record_probe` or
            :py:meth:`navground.sim.Experiment.add_record_probe`.
        """
        probe = functools.partial(cls, indices=indices)
        return cast('Callable[[sim.Dataset], SuccessProbe]', probe)

    @classmethod
    def with_criteria(
        cls,
        is_success: Condition | None = None,
        is_failure: Condition | None = None
    ) -> Callable[[sim.Dataset], SuccessProbe]:
        probe = functools.partial(cls,
                                  is_success=is_success,
                                  is_failure=is_failure)
        return cast('Callable[[sim.Dataset], SuccessProbe]', probe)



from __future__ import annotations

import numpy as np
from navground import core, sim


class AdvanceTask(sim.Task, name="Advance"):

    def __init__(self,
                 direction: core.Vector2 = core.unit(0),
                 target: float = 1.0) -> None:
        super().__init__()
        self.direction = direction / np.linalg.norm(direction)
        self.target = target
        self._advancement = 0.0

    def update(self, agent: sim.Agent, world: sim.World, time: float) -> None:
        self._advancement = np.dot(
            agent.pose.position - self._initial_position, self._direction)
        if self.done():
            agent.controller.stop()

    def prepare(self, agent: sim.Agent, world: sim.World) -> None:
        self._initial_position = np.array(agent.pose.position)
        agent.controller.follow_velocity(self.direction)

    def done(self) -> bool:
        return self._advancement >= self._target_advancement

    @property
    @sim.register(core.unit(0), "direction")
    def direction(self) -> core.Vector2:
        return self._direction

    @direction.setter
    def direction(self, value: core.Vector2) -> None:
        norm = np.linalg.norm(value)
        if norm > 0:
            self._direction = value / norm

    @property
    @sim.register(1.0, "target advancement")
    def target(self) -> float:
        return self._target_advancement

    @target.setter
    def target(self, value: float) -> None:
        self._target_advancement = max(0, value)



from __future__ import annotations

from . import env, parallel_env, rewards
from .config import (ActionConfig, ControlActionConfig,
                     DefaultObservationConfig, DefaultStateConfig, GroupConfig,
                     ModulationActionConfig, ObservationConfig, StateConfig)
from .indices import Indices

__all__ = [
    'ControlActionConfig', 'DefaultObservationConfig', 'ObservationConfig',
    'ModulationActionConfig', 'GroupConfig', 'ActionConfig', 'rewards', 'env',
    'parallel_env', 'Indices', 'StateConfig', 'DefaultStateConfig'
]



from __future__ import annotations

from torchrl.envs.libs.pettingzoo import PettingZooWrapper  # type: ignore

from benchmarl.experiment import Experiment  # type: ignore[import-not-found]

from ...parallel_env import MultiAgentNavgroundEnv
from ...types import PathLike
from ...wrappers.name_wrapper import NameWrapper
from .callbacks import AlternateActorCallback, ExportPolicyCallback
from .evaluate import evaluate_policy
from .navground_experiment import NavgroundExperiment
from .policy import SingleAgentPolicy
from .split_mlp import SplitMlpConfig
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from torchrl.envs import EnvBase  # type: ignore

def make_env(env: MultiAgentNavgroundEnv,
             seed: int = 0,
             categorical_actions: bool = False) -> EnvBase:
    return PettingZooWrapper(NameWrapper(MultiAgentNavgroundEnv(**env._spec)),
                             categorical_actions=categorical_actions,
                             device='cpu',
                             seed=seed,
                             return_state=env.has_state)


def reload_experiment(path: PathLike) -> Experiment:
    return Experiment.reload_from_file(str(path))


__all__ = [
    'ExportPolicyCallback', 'AlternateActorCallback', 'NavgroundExperiment',
    'SingleAgentPolicy', 'make_env', 'evaluate_policy', 'reload_experiment',
    'SplitMlpConfig'
]



from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import torch

if TYPE_CHECKING:
    from benchmarl.models.mlp import Mlp  # type: ignore[import-not-found]

from ...types import PyTorchObs


class MlpPolicy(torch.nn.Module):

    def __init__(self, model: Mlp):
        super().__init__()
        mlp = model.mlp
        self.mlp = copy.deepcopy(mlp._empty_net)
        mlp.params.to_module(self.mlp)
        self.in_keys = [key[-1] for key in model.in_keys]

    def forward(self, observation: PyTorchObs) -> torch.Tensor:
        if isinstance(observation, dict):
            x: torch.Tensor = torch.cat(
                [observation[key] for key in self.in_keys], dim=-1)
        else:
            x = observation
        return self.mlp(x)



from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING, Any

from .base import BaseILAlgorithm

if TYPE_CHECKING:
    from collections.abc import Mapping

    import gymnasium as gym
    from imitation.util.logger import HierarchicalLogger
    from stable_baselines3.common.vec_env import VecEnv

    from .rollout import AnyPolicy


class DAgger(BaseILAlgorithm):
    """
    A simplified interface to out version of
    :py:class:`imitation.algorithms.dagger.SimpleDAggerTrainer`
    that accepts :py:type:`PolicyCallableWithInfo` experts.

    :param env: the environment.
    :param seed: the random seed
    :param policy: an optional policy
    :param policy_kwargs: or the kwargs to create it
    :param logger: an optional logger
    :param expert: the expert to imitate
    :param bc_kwargs: parameters passed to the
                      :imitation.algorithms.bc.BC: constructor
    :param dagger_kwargs: parameters passed to the
        :py:class:`imitation.algorithms.dagger.SimpleDAggerTrainer` constructor
    """

    def __init__(self,
                 env: gym.Env[Any, Any] | VecEnv | None = None,
                 seed: int = 0,
                 policy: Any = None,
                 policy_kwargs: Mapping[str, Any] = {'net_arch': [32, 32]},
                 logger: HierarchicalLogger | None = None,
                 expert: AnyPolicy | None = None,
                 bc_kwargs: Mapping[str, Any] = {},
                 dagger_kwargs: dict[str, Any] = {}):
        from imitation.algorithms import bc

        from .dagger_trainer import SimpleDAggerTrainer

        self._temp_dir: tempfile.TemporaryDirectory[Any] | None = None
        super().__init__(env=env,
                         seed=seed,
                         policy=policy,
                         policy_kwargs=policy_kwargs,
                         logger=logger)
        self._expert = expert
        if not expert and self.env:
            try:
                self._expert = self.env.get_attr("policy", [0])[0]
            except (AttributeError, IndexError):
                pass
        self._bc_kwargs = bc_kwargs
        self._dagger_kwargs = dagger_kwargs
        self._bc_trainer = bc.BC(observation_space=self.observation_space,
                                 action_space=self.action_space,
                                 rng=self.rng,
                                 policy=self._policy,
                                 **self._bc_kwargs)
        if self.env and self._expert:
            self._temp_dir = tempfile.TemporaryDirectory(prefix="dagger")
            self._trainer = SimpleDAggerTrainer(
                venv=self.env,
                scratch_dir=self._temp_dir.name,
                expert_policy=self._expert,  # type: ignore[arg-type]
                bc_trainer=self._bc_trainer,
                rng=self.rng,
                **self._dagger_kwargs)
            self.set_logger(self.logger)

    def set_logger(self, logger: HierarchicalLogger) -> None:
        super().set_logger(logger)
        if self._bc_trainer:
            self._bc_trainer.logger = logger
            self._bc_trainer._bc_logger._logger = logger

    def __del__(self) -> None:
        if self._temp_dir:
            self._temp_dir.cleanup()



from __future__ import annotations

import numpy as np
from navground import sim


class CommProbe(sim.RecordProbe):
    """
    A probe that records transmitted messages to a single dataset
    of shape ``(#agents, message size)``
    """

    dtype = np.float32

    def update(self, run: sim.ExperimentalRun) -> None:
        for agent in run.world.agents:
            try:
                comm = agent.behavior._comm  # type: ignore[union-attr]
            except AttributeError:
                continue
            self.data.append(comm)

    def get_shape(self, world: sim.World) -> list[int]:
        shapes = []
        for agent in world.agents:
            try:
                shapes.append(agent.behavior.  # type: ignore[union-attr]
                              _gym_agent.action_config.comm_space.shape)
            except AttributeError:
                continue
        assert len(set(shapes)) == 1
        return [len(shapes), *shapes[0]]



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

try:
    from typing import Self
except ImportError:
    try:
        from typing_extensions import Self
    except ImportError:
        pass


class Registrable:
    """
    Internal helper class that can be converted to/from a plain dictionary
    """

    _type: str = ''

    @property
    def asdict(self) -> dict[str, Any]:
        """
        The YAML-able dictionary representation.

        Adds the class type name to the dictionary returned by
        :py:meth:`_get_dict`.

        :returns: A YAML-able dictionary
        """
        return dict(**self._get_dict(), type=self._type)

    @classmethod
    def from_dict(cls, value: Mapping[str, Any]) -> Self:
        """
        Load an instance from a dictionary.

        Selects a class from field ``type`` and calls its method
        :py:meth:`_make_from_dict`.

        :param      value:  The value

        :raises ValueError: If ``type`` is not defined or not registered.

        :returns:   An instance initialized from the content of ``value``.
        """
        if 'type' not in value:
            raise ValueError(f"{cls.__name__}: type not provided")
        t = value.get('type')
        for subcls in cls.__subclasses__():
            if subcls._type == t:
                vs = {k: v for k, v in value.items() if k != 'type'}
                return subcls._make_from_dict(vs)
        raise ValueError(
            f"{cls.__name__}: type {value['type']} not registered")

    def _get_dict(self) -> dict[str, Any]:
        """
        Sub-classes can override this method to
        provide fields in :py:data:`asdict`.

        The default implementation returns an empty dict.

        :returns: A YAML-able dictionary
        """
        return {}

    @classmethod
    def _make_from_dict(cls, value: Mapping[str, Any]) -> Self:
        """
        Sub-classes can override this method to load an instance from
        a dictionary.

        The default implementation initializes the class with no arguments.

        :param      value:  The value

        :returns:   An instance initialized from the content of ``value``.
        """
        return cls()

    # Renamed to `register_name` because `name` conflicts with Python3.10 ABC
    def __init_subclass__(cls, register_name: str = '') -> None:
        cls._type = register_name



from __future__ import annotations

from collections.abc import Collection, Mapping
from typing import TYPE_CHECKING, Any, TypeAlias

import numpy as np
from navground import sim
from pettingzoo.utils.env import ParallelEnv

from ..config import (ActionConfig, ControlActionConfig,
                      DefaultObservationConfig, GroupConfig, ObservationConfig,
                      StateConfig)
from ..env import BaseEnv, NavgroundEnv
from ..indices import Indices, IndicesLike
from ..internal.base_env import NavgroundBaseEnv, ResetReturn, StepReturn
from ..rewards import NullReward
from ..types import (Action, Array, Bounds, Observation, Reward, SensorLike,
                     SensorSequenceLike, TerminationCondition)

if TYPE_CHECKING:
    import gymnasium as gym

BaseParallelEnv: TypeAlias = ParallelEnv[int, Observation, Action]


class MultiAgentNavgroundEnv(NavgroundBaseEnv, BaseParallelEnv):
    """
    This class describes an environment that uses
    a :py:class:`navground.sim.Scenario` to
    generate and then simulate a :py:class:`navground.sim.World`.

    Actions and observations relates to one or more selected
    :py:class:`navground.sim.Agent`.

    We provide convenience functions to initialize the class
    :py:func:`parallel_env` and :py:func:`shared_parallel_env`
    to create the environment.
    """

    metadata = {
        "name": "navground_ma",
        "render_modes": ["human", "rgb_array"],
        'render_fps': 30
    }
    MAX_SEED = 2**31 - 1

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        NavgroundBaseEnv.__init__(self, *args, **kwargs)
        self.possible_agents = list(self._possible_agents.keys())
        self.agents = list(self._agents.keys())

    def reset(self,
              seed: int | None = None,
              options: dict[str, Any] | None = None) -> ResetReturn:
        """
        Conforms to :py:meth:`pettingzoo.utils.env.ParallelEnv.reset`.

        It samples a new world from a scenario, runs one dry simulation step
        using :py:meth:`navground.sim.World.update_dry`.
        Then, it converts the agents states to observations.
        and the commands they would actuate to actions, which it includes
        in the ``info`` dictionary at key `"navground_action"``.
        """
        NavgroundBaseEnv._reset(self, seed, options)
        self.agents = list(self._agents.keys())
        return self.get_observations(), self.get_infos()

    def step(self, actions: dict[int, Action]) -> StepReturn:
        """
        Conforms to :py:meth:`pettingzoo.utils.env.ParallelEnv.step`.

        It converts the actions to commands that the navground agents actuate.
        Then, it updates the world for one step,
        calling :py:meth:`navground.sim.World.update`.
        Finally, it converts the agents states to observations,
        the commands they would actuate to actions, which it includes
        in the ``info`` dictionary at key `"navground_action"`, and computes rewards.

        Termination for individual agents is set when they complete the task,
        exit the boundary, or get stuck.
        Truncation for all agents is set when the maximal duration has passed.
        """
        obs, rew, term, trunc, info = NavgroundBaseEnv._step(self, actions)
        self.agents = [
            index for index in self.agents
            if not (trunc.get(index, False) or term.get(index, False))
        ]
        return obs, rew, term, trunc, info

    def observation_space(self, agent: int) -> gym.Space[Any]:
        return self._observation_space[agent]

    def action_space(self, agent: int) -> gym.Space[Any]:
        return self._action_space[agent]

    def state(self) -> Array:
        vs = self.get_state()
        if vs is None:
            vs = np.empty((0, ), dtype=np.float32)
        return vs


def parallel_env(
    scenario: sim.Scenario | str | dict[str, Any] | None = None,
    groups: Collection[GroupConfig] = tuple(),
    max_number_of_agents: int | None = None,
    time_step: float = 0.1,
    max_duration: float = -1.0,
    terminate_if_idle: bool = True,
    bounds: Bounds | None = None,
    truncate_outside_bounds: bool = False,
    render_mode: str | None = None,
    render_kwargs: Mapping[str, Any] = {},
    realtime_factor: float = 1.0,
    stuck_timeout: float = -1,
    wait: bool = False,
    truncate_fast: bool = False,
    state: StateConfig | None = None,
    include_action: bool = True,
    include_success: bool = True,
    init_success: bool | None = None,
    intermediate_success: bool = False,
) -> MultiAgentNavgroundEnv:
    """
    Create a multi-agent PettingZoo environment that uses
    a :py:class:`navground.sim.Scenario` to
    generate and then simulate a :py:class:`navground.sim.World`.


    >>> from navground.learning.parallel_env import parallel_env
    >>> from navground.learning import GroupConfig
    >>> from navground import sim
    >>>
    >>> scenario = sim.load_scenario(...)
    >>> groups = [GroupConfig(...), ...]
    >>> env = parallel_env(scenario=scenario, groups=groups)


    :param scenario: The scenario to initialize all simulated worlds.
                     If a :py:class:`str`, it will be interpreted as the YAML
                     representation of a scenario.
                     If a :py:class:`dict`, it will be dumped to YAML and
                     then treated as a :py:class:`str`.

    :param groups: The configuration of the agents controlled by the environment.
                   All other agents are controlled solely by navground.

    :param max_number_of_agents: The maximal number of agents that we will expose.
                                 It needs to be specified only for scenarios
                                 that generate world with a variable number of agents.

    :param time_step: The simulation time step applied
                      at every :py:meth:`MultiAgentNavgroundEnv.step`.

    :param max_duration: If positive, it will signal a truncation after this simulated time.

    :param terminate_if_idle: Whether to terminate when an agent is idle

    :param truncate_outside_bounds: Whether to truncate when an agent exit the bounds

    :param bounds: The area to render and a fence for truncating processes when agents exit it.

    :param render_mode: The render mode.
                        If `"human"`, it renders a simulation in real time via
                        websockets (see :py:class:`navground.sim.ui.WebUI`).
                        If `"rgb_array"`, it uses
                        :py:func:`navground.sim.ui.render.image_for_world`
                        to render the world on demand.

    :param render_kwargs: Arguments passed to :py:func:`navground.sim.ui.render.image_for_world`

    :param realtime_factor: a realtime factor for `render_mode="human"`: larger values
                            speed up the simulation.

    :param stuck_timeout: The time to wait before considering an agent stuck.

    :param wait: Whether to signal termination/truncation only when
                 all agents have terminated/truncated.

    :param truncate_fast: Whether to signal truncation for all agents
                          as soon as one agent truncates.

    :param state: An optional global state configuration

    :param include_action: Whether to include field "navground_action" in the info

    :param include_success: Whether to include field "is_success" in the info

    :param init_success: The default value of success (valid until a termination condition is met)

    :param intermediate_success: Whether to include "is_success" in the info at intermediate
                                 steps (vs only at termination).

    :returns:   The multi agent navground environment.
    """
    return MultiAgentNavgroundEnv(
        groups=groups,
        scenario=scenario,
        max_number_of_agents=max_number_of_agents,
        time_step=time_step,
        max_duration=max_duration,
        terminate_if_idle=terminate_if_idle,
        bounds=bounds,
        truncate_outside_bounds=truncate_outside_bounds,
        render_mode=render_mode,
        render_kwargs=render_kwargs,
        realtime_factor=realtime_factor,
        stuck_timeout=stuck_timeout,
        wait=wait,
        truncate_fast=truncate_fast,
        include_action=include_action,
        include_success=include_success,
        init_success=init_success,
        intermediate_success=intermediate_success,
        state=state)


def shared_parallel_env(
        scenario: sim.Scenario | str | dict[str, Any] | None = None,
        max_number_of_agents: int | None = None,
        indices: IndicesLike = Indices.all(),
        sensor: SensorLike | None = None,
        sensors: SensorSequenceLike = tuple(),
        action: ActionConfig = ControlActionConfig(),
        observation: ObservationConfig = DefaultObservationConfig(),
        reward: Reward | None = None,
        time_step: float = 0.1,
        max_duration: float = -1.0,
        terminate_if_idle: bool = True,
        bounds: Bounds | None = None,
        truncate_outside_bounds: bool = False,
        render_mode: str | None = None,
        render_kwargs: Mapping[str, Any] = {},
        realtime_factor: float = 1.0,
        stuck_timeout: float = -1,
        color: str = '',
        tag: str = '',
        wait: bool = False,
        truncate_fast: bool = False,
        terminate_on_success: bool = True,
        terminate_on_failure: bool = True,
        success_condition: TerminationCondition | None = None,
        failure_condition: TerminationCondition | None = None,
        include_action: bool = True,
        include_success: bool = True,
        init_success: bool | None = None,
        intermediate_success: bool = False,
        state: StateConfig | None = None) -> MultiAgentNavgroundEnv:
    """
    Create a multi-agent PettingZoo environment that uses
    a :py:class:`navground.sim.Scenario` to
    generate and then simulate a :py:class:`navground.sim.World`.
    All controlled agents share the same configuration.


    >>> from navground.learning.parallel_env import shared_parallel_env
    >>> from navground import sim
    >>>
    >>> scenario = sim.load_scenario(...)
    >>> env = shared_parallel_env(scenario=scenario, ...)

    :param scenario: The scenario to initialize all simulated worlds.
                     If a :py:class:`str`, it will be interpreted as the YAML
                     representation of a scenario.
                     If a :py:class:`dict`, it will be dumped to YAML and
                     then treated as a :py:class:`str`.

    :param max_number_of_agents: The maximal number of agents that we will expose.
                                 It needs to be specified only for scenarios
                                 that generate world with a variable number of agents.

    :param indices: The world indices of the agent to control.
                    All other agents are controlled solely by navground.

    :param sensor: An optional sensor that will be added to ``sensors``.

    :param sensors: A sequence of sensor to generate observations for the agents
                    or its YAML representation. If
                    Items of class :py:class:`str` will be interpreted as the YAML
                    representation of a sensor.
                    Items of class :py:class:`dict` will be dumped to YAML and
                    then treated as a :py:class:`str`.
                    If empty, it will use the agents' own sensors.

    :param action: The configuration of the action space to use.

    :param observation: The configuration of the observation space to use.

    :param reward: The reward function to use. If none, it will default to constant zeros.

    :param time_step: The simulation time step applied
                      at every :py:meth:`MultiAgentNavgroundEnv.step`.

    :param max_duration: If positive, it will signal a truncation after this simulated time.

    :param terminate_if_idle: Whether to terminate when an agent is idle

    :param truncate_outside_bounds: Whether to truncate when an agent exit the bounds

    :param bounds: The area to render and a fence for truncating processes when agents exit it.

    :param render_mode: The render mode.
                        If `"human"`, it renders a simulation in real time via
                        websockets (see :py:class:`navground.sim.ui.WebUI`).
                        If `"rgb_array"`, it uses
                        :py:func:`navground.sim.ui.render.image_for_world`
                        to render the world on demand.

    :param render_kwargs: Arguments passed to :py:func:`navground.sim.ui.render.image_for_world`

    :param realtime_factor: a realtime factor for `render_mode="human"`: larger values
                            speed up the simulation.

    :param stuck_timeout: The time to wait before considering an agent stuck.

    :param color: An optional color of the agents (only used for displaying)

    :param tag: An optional tag to be added to the agents (only used as metadata)

    :param wait: Whether to signal termination/truncation only when
                 all agents have terminated/truncated.

    :param truncate_fast: Whether to signal truncation for all agents
                          as soon as one agent truncates.

    :param state: An optional global state configuration

    :param terminate_on_success: Whether to terminate the episode on success.

    :param terminate_on_failure: Whether to terminate the episode on failure.

    :param success_condition: An optional success criteria

    :param failure_condition: An optional failure criteria

    :param include_action: Whether to include field "navground_action" in the info

    :param include_success: Whether to include field "is_success" in the info

    :param init_success: The default value of success (valid until a termination condition is met)

    :param intermediate_success: Whether to include "is_success" in the info at intermediate
                                 steps (vs only at termination).



    :returns:   The multi agent navground environment.
    """
    if reward is None:
        reward = NullReward()
    group = GroupConfig(indices=Indices(indices),
                        sensor=sensor,
                        sensors=sensors,
                        reward=reward,
                        action=action,
                        observation=observation,
                        color=color,
                        tag=tag,
                        terminate_on_success=terminate_on_success,
                        terminate_on_failure=terminate_on_failure,
                        success_condition=success_condition,
                        failure_condition=failure_condition)
    return MultiAgentNavgroundEnv(
        groups=[group],
        scenario=scenario,
        time_step=time_step,
        max_duration=max_duration,
        terminate_if_idle=terminate_if_idle,
        bounds=bounds,
        truncate_outside_bounds=truncate_outside_bounds,
        render_mode=render_mode,
        render_kwargs=render_kwargs,
        realtime_factor=realtime_factor,
        stuck_timeout=stuck_timeout,
        wait=wait,
        truncate_fast=truncate_fast,
        include_action=include_action,
        include_success=include_success,
        init_success=init_success,
        intermediate_success=intermediate_success,
        state=state)


def make_shared_parallel_env_with_env(
    env: BaseEnv,
    max_number_of_agents: int | None = None,
    indices: IndicesLike = Indices.all()
) -> MultiAgentNavgroundEnv:
    """
    Creates a shared parallel environment
    using the configuration of the agent exposed
    in a (navground) single-agent environment.


    :param      env:                   The environment.
    :param      max_number_of_agents:  The maximal number of agents that we will expose.
        It needs to be specified only for scenarios
        that generate world with a variable number of agents.
    :param      indices:         The world indices of the agent to control.
        All other agents are controlled solely by navground.

    :raises TypeError: If ``env.unwrapped`` is not a subclass of :py:class:`.env.NavgroundEnv`

    """
    if not isinstance(env.unwrapped, NavgroundEnv):
        raise TypeError("Environment is not a NavgroundEnv")

    kwargs = dict(**env.unwrapped._spec)
    kwargs.pop('agent_index')
    kwargs.pop('reward')
    kwargs.pop('action')
    kwargs.pop('observation')
    kwargs.pop('sensor')
    kwargs.pop('sensors')
    kwargs['max_number_of_agents'] = max_number_of_agents
    group = list(env.unwrapped.groups_config)[0]
    group.indices = Indices(indices)
    return MultiAgentNavgroundEnv(groups=[group], **kwargs)



import gymnasium as gym
from navground import sim

from ..config import ControlActionConfig, DefaultObservationConfig
from ..env import BaseEnv
from ..rewards import SocialReward
from ..scenarios import CorridorWithObstacle  # noqa: F401


def get_env(flat: bool = True,
            duration: float = 40.0,
            time_step: float = 0.1) -> BaseEnv:
    """
    Creates the an environment where a single
    agent traveling along a corridor with a single static obstacle.

    :param flat: Whether the observation space is flat
    :param duration: The duration of an episode
    :param time_step: The simulation time step

    :returns: A Gymnasium environment

    """

    scenario = sim.load_scenario("""
type: CorridorWithObstacle
length: 1.0
width: 1.0
min_radius: 0.2
max_radius: 0.2
groups:
  -
    type: thymio
    number: 1
    radius: 0.08
    control_period: 0.05
    color: blue
    kinematics:
      type: 2WDiff
      wheel_axis: 0.094
      max_speed: 0.12
    behavior:
      type: HL
      optimal_speed: 0.12
      horizon: 10
      tau: 0.25
      eta: 0.5
      safety_margin: 0.05
      barrier_angle: 1.0
    state_estimation:
      type: Bounded
      range: 1.0
      update_static_obstacles: true
""")

    sensors = """
type: Combination
sensors:
  - type: Boundary
    min_y: 0
    max_y: 1
    range: 1
  - type: Discs
    number: 1
    range: 1
    max_speed: 0.0
    max_radius: 0.0
    include_valid: false
"""

    action_config = ControlActionConfig(max_acceleration=1.0,
                                        max_angular_acceleration=10.0,
                                        use_acceleration_action=True)

    observation_config = DefaultObservationConfig(
        include_target_distance=False,
        include_target_direction=True,
        include_velocity=True,
        include_angular_speed=True,
        flat=flat)
    return gym.make('navground',
                    scenario=scenario,
                    sensors=sensors,
                    action=action_config,
                    observation=observation_config,
                    time_step=time_step,
                    max_duration=duration,
                    reward=SocialReward(safety_margin=0.04))



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pettingzoo.utils.wrappers import BaseParallelWrapper

from ..internal.base_env import ResetReturn, StepReturn
from ..parallel_env import MultiAgentNavgroundEnv
from ..types import Action

if TYPE_CHECKING:
    from pettingzoo.utils.env import ParallelEnv


class NameWrapper(BaseParallelWrapper):
    """
    This wrapper renames the agent using their *navground* tags
    as prefix "tag1-tag2-...-tagn" or "agent" (if no tag is set).

    Indices are appended to the predix, e.g.
    agent_0, agent_1, ... .

    :param env: The wrapped environment
    """

    def __init__(self, env: ParallelEnv) -> None:
        assert isinstance(env.unwrapped, MultiAgentNavgroundEnv)
        super().__init__(env)
        self._names: dict[int, str] = {
            i: f'{"-".join(agent.navground.tags) or "agent"}_{i}'
            for i, agent in env.unwrapped._possible_agents.items()
            if agent.navground
        }
        self._original_names: dict[str, int] = {
            v: k
            for k, v in self._names.items()
        }
        self.possible_agents = list(self._original_names)


    def get_indices(self, group: str) -> list[int]:
        """
        Gets the indices of agents with name <group>_<index>

        :param      group:  The group

        :returns:   The agent indices in the wrapped environment
        """
        return [
            i for name, i in self._original_names.items() if
            name.startswith(group)
        ]

    @property
    def agents(self) -> list[str]:
        return [self._names[i] for i in self.env.agents]

    @agents.setter
    def agents(self, values: list[str]) -> None:
        self.env.agents = [
            self._original_names[i] for i in values
            if i in self._original_names
        ]

    def observation_space(self, agent):
        return self.env.observation_space(self._original_names[agent])

    def action_space(self, agent):
        return self.env.action_space(self._original_names[agent])

    def reset(self,
              seed: int | None = None,
              options: dict[str, Any] | None = None) -> ResetReturn:

        obs, infos = self.env.reset(seed=seed, options=options)
        obs = {self._names[k]: v for k, v in obs.items()}
        infos = {self._names[k]: v for k, v in infos.items()}
        return obs, infos

    def step(self, action: dict[str, Action]) -> StepReturn:
        r_action = {self._original_names[k]: v for k, v in action.items()}
        obs, reward, terminated, truncated, infos = self.env.step(r_action)
        obs = {self._names[k]: v for k, v in obs.items()}
        reward = {self._names[k]: v for k, v in reward.items()}
        terminated = {self._names[k]: v for k, v in terminated.items()}
        truncated = {self._names[k]: v for k, v in truncated.items()}
        infos = {self._names[k]: v for k, v in infos.items()}
        return obs, reward, terminated, truncated, infos



from __future__ import annotations

import datetime
import pathlib as pl
import warnings
from typing import TYPE_CHECKING, Any

try:
    from typing import Self
except ImportError:
    try:
        from typing_extensions import Self
    except ImportError:
        ...

import gymnasium as gym
import numpy as np

from ..types import PathLike
from .utils import maybe_make_venv

# from stable_baselines3.common.callbacks import BaseCallback
# from stable_baselines3.common.logger import Logger, configure


if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from imitation.util.logger import HierarchicalLogger
    from stable_baselines3.common.policies import ActorCriticPolicy
    from stable_baselines3.common.vec_env import VecEnv

ExtractorFactory = Any

ISO_TIMESTAMP = "%Y%m%d_%H%M%S"


def make_policy(
        observation_space: gym.Space[Any],
        action_space: gym.Space[Any],
        net_arch: list[int] = [32, 32],
        features_extractor: ExtractorFactory | None = None,
        features_extractor_kwargs: dict[str, Any] = {}) -> ActorCriticPolicy:
    import torch
    from stable_baselines3.common import torch_layers
    from stable_baselines3.common.policies import ActorCriticPolicy

    if features_extractor is None:
        if isinstance(observation_space, gym.spaces.Dict):
            features_extractor = torch_layers.CombinedExtractor
        else:
            features_extractor = torch_layers.FlattenExtractor
    return ActorCriticPolicy(
        observation_space=observation_space,
        action_space=action_space,
        lr_schedule=lambda _: torch.finfo(torch.float32).max,
        features_extractor_class=features_extractor,
        features_extractor_kwargs=features_extractor_kwargs,
        net_arch=net_arch)


def make_logger(log_directory: PathLike = "logs",
                stamp_log_directory: bool = True,
                log_formats: Sequence[str] = []) -> HierarchicalLogger:
    from imitation.util.logger import configure

    log_directory = pl.Path(log_directory)
    if any(fmt != 'stdout' for fmt in log_formats):
        if stamp_log_directory:
            timestamp = datetime.datetime.now().strftime(ISO_TIMESTAMP)
            log_directory = log_directory / timestamp
    return configure(log_directory, log_formats)


class BaseILAlgorithm:
    """
    The base class that wraps IL algorithms to implement a
    :py:class:`stable_baselines3.common.base_class.BaseAlgorithm`-like API.

    :param env: the environment.
    :param seed: the random seed
    :param policy: an optional policy
    :param policy_kwargs: or the kwargs to create it
    :param logger: an optional logger
    """

    def __init__(self,
                 env: gym.Env[Any, Any] | VecEnv | None = None,
                 seed: int = 0,
                 policy: Any = None,
                 policy_kwargs: Mapping[str, Any] = {'net_arch': [32, 32]},
                 logger: HierarchicalLogger | None = None):
        if logger is None:
            logger = make_logger()
        self._logger = logger
        self._trainer: Any = None
        self.rng = np.random.default_rng(seed)
        if env:
            self._env: VecEnv | None = maybe_make_venv(env=env, rng=self.rng)
            self._observation_space = self._env.observation_space
            self._action_space = self._env.action_space
        else:
            self._env = None
        if policy:
            self._policy = policy
            if self._env:
                if (policy.action_space != self._env.action_space
                        or policy.observation_space
                        != self._env.observation_space):
                    raise ValueError("Policy and env have different spaces")
            else:
                self._observation_space = policy.observation_space
                self._action_space = policy.action_space
        else:
            if self._env:
                self._policy = make_policy(self._env.observation_space,
                                           self._env.action_space,
                                           **policy_kwargs)
            else:
                raise ValueError(
                    "At least one of env or policy must be not None")

    @property
    def action_space(self) -> gym.Space[Any]:
        """The action space"""
        return self._action_space

    @property
    def observation_space(self) -> gym.Space[Any]:
        """The observation space"""
        return self._observation_space

    @property
    def policy(self) -> Any:
        """
        Gets the policy
        """
        return self._policy

    @property
    def env(self) -> VecEnv | None:
        """The training env"""
        return self.get_env()

    @env.setter
    def env(self, env: gym.Env[Any, Any] | VecEnv) -> None:
        self.set_env(env)

    def set_logger(self, logger: HierarchicalLogger) -> None:
        """
        Sets the logger.

        :param      logger:  The logger
        """
        self._logger = logger
        if self._trainer:
            self._trainer.logger = logger

    def get_env(self) -> VecEnv | None:
        """
        Gets the training environment.

        :returns:   The environment.
        """
        return self._env

    def set_env(self, env: gym.Env[Any, Any] | VecEnv) -> None:
        """
        Sets the training environment.

        Rejects the enviroment if not compatible with the policy.

        :param  env:  The new environment
        """
        if (self.policy.action_space == env.action_space
                and self.policy.observation_space == env.observation_space):
            self._env = maybe_make_venv(env)
        else:
            warnings.warn("Environment is not compatible with policy",
                          stacklevel=1)

    @property
    def logger(self) -> HierarchicalLogger:
        """
        The logger
        """
        return self._logger

    @logger.setter
    def logger(self, logger: HierarchicalLogger) -> None:
        self.set_logger(logger)

    def learn(
        self,
        *args: Any,
        # callback: Callable | BaseCallback | list[BaseCallback] | None = None,
        **kwargs: Any
    ) -> Self:
        """
        Learns the policy, passing the arguments to the wrapped ``imitation`` trainer.
        like

        .. code-block:: python

           trainer.train(*args, **kwargs)

        :returns: self
        """

        # if callback:
        #     callback.init_callback_for_imitation(self)
        #     kwargs['on_batch_end'] = lambda steps: callback.step(steps=steps)
        if self._trainer:
            self._trainer.train(*args, **kwargs)
        else:
            warnings.warn("Missing trainer", stacklevel=1)
        return self

    def save(self, path: pl.Path | str) -> None:
        """
        Saves the model using :py:func:`stable_baselines3.common.save_util.save_to_zip_file`

        :param      path:  The path to the directory where to create
        """
        from stable_baselines3.common.save_util import save_to_zip_file

        data = {'policy_kwargs': self.policy._get_constructor_parameters()}
        save_to_zip_file(save_path=path,
                         data=data,
                         params={'policy': self.policy.state_dict()},
                         pytorch_variables={})

    @classmethod
    def load(cls, path: pl.Path | str, env: gym.Env[Any, Any] | VecEnv | None = None) -> Self:
        """
        Loads a model using :py:func:`stable_baselines3.common.save_util.load_from_zip_file`

        :param      path:  The path to the saved model
        :param      env:   An optional training environment

        :returns:   the loaded algorithm
        """
        import torch
        from stable_baselines3.common.policies import ActorCriticPolicy
        from stable_baselines3.common.save_util import load_from_zip_file

        torch.serialization.add_safe_globals([ActorCriticPolicy])

        data, params, pytorch_variables = load_from_zip_file(path)
        if data:
            policy_kwargs: dict[str, Any] = data.get('policy_kwargs', {})
        else:
            policy_kwargs = {}
        policy = ActorCriticPolicy(**policy_kwargs)
        policy.load_state_dict(params['policy'])  # type: ignore[arg-type]
        return cls(env=env, policy=policy)



from __future__ import annotations

import copy
from collections.abc import Collection
from functools import partial
from typing import TYPE_CHECKING

from navground import sim

from ..config import GroupConfig, merge_groups_configs
from ..env import BaseEnv
from ..indices import join_indices
from ..internal.base_env import NavgroundBaseEnv
from ..parallel_env import BaseParallelEnv
from ..probes.reward import RewardProbe
from ..probes.success import SuccessProbe
from ..types import (AnyPolicyPredictor, Bounds, ObservationTransform,
                     PathLike, Reward)
from .scenario import InitPolicyBehavior

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv


def make_experiment(scenario: sim.Scenario,
                    groups: Collection[GroupConfig] = tuple(),
                    reward: Reward | None = None,
                    record_reward: bool = True,
                    policy: AnyPolicyPredictor | PathLike = '',
                    bounds: Bounds | None = None,
                    terminate_outside_bounds: bool = True,
                    deterministic: bool = True) -> sim.Experiment:
    """
    Initializes an navground experiment where groups of agents
    are configured with possibly different policies and
    the rewards is optionally recorded.

    If ``groups`` is not empty, it make a copy of the scenario, to which it adds
    adds :py:class:`navground.learning.evaluation.scenario.InitPolicyBehavior`
    to initialize the groups.

    If ``record_reward`` is set, it adds a
    :py:class:`navground.learning.probes.reward.RewardProbe`.

    :param  scenario:                  The scenario
    :param  groups:                    The configuration of the groups
    :param  reward:                    The default reward to record
                                       (when not specified in the group config)
    :param  record_reward:             Whether to record the rewards
    :param  policy:                    The default policy
                                       (when not specified in the group config)
    :param  bounds:                    Optional termination boundaries
    :param  terminate_outside_bounds:  Whether to terminate
                                       if some of the agents exits the boundaries
    :param  deterministic:             Whether to apply the policies deterministically

    :returns:   The experiment
    """
    experiment = sim.Experiment()
    experiment.terminate_when_all_idle_or_stuck = False
    if not groups:
        groups = [GroupConfig(policy=policy)]
    if groups:
        experiment.scenario = copy.deepcopy(scenario)
        init = InitPolicyBehavior(
            groups=groups,
            bounds=bounds,
            terminate_outside_bounds=terminate_outside_bounds,
            deterministic=deterministic)
        experiment.scenario.add_init(init)
    else:
        experiment.scenario = scenario
    if record_reward:
        experiment.add_record_probe(
            "reward", partial(RewardProbe, groups=groups, reward=reward))
    return experiment


def make_experiment_with_env(env: BaseEnv | BaseParallelEnv | VecEnv,
                             groups: Collection[GroupConfig] = tuple(),
                             policy: AnyPolicyPredictor | PathLike = '',
                             reward: Reward | None = None,
                             record_reward: bool = True,
                             record_success: bool = True,
                             deterministic: bool = True,
                             grouped: bool = False,
                             pre: ObservationTransform | None = None) -> sim.Experiment:
    """
    Similar to :py:func:`make_experiment` but using the configuration stored in
    an environment:
    ``groups`` are merged using :py:func:`navground.learning.config.merge_groups_configs`.


    :param  env:                       The environment
    :param  groups:                    The configuration of the groups
    :param  reward:                    The default reward to record
                                       (when not specified in the group config)
    :param  record_reward:             Whether to record the rewards
    :param  record_success:             Whether to record the success
    :param  policy:                    The default policy
                                       (when not specified in the group config)
    :param  grouped:                   Whether the policy is grouped.
    :param  deterministic:             Whether to apply the policies deterministically
    :param  pre:                       An optional transformation to apply to observations

    :returns:   The experiment
    """
    from stable_baselines3.common.vec_env import VecEnv

    if not groups:
        groups = [GroupConfig(policy=policy)]
    # Should not be anymore necessary as `get_attr` is added in `make_vec_from_penv`
    # if isinstance(env, VecEnv):
    #     try:
    #         env = env.unwrapped.vec_envs[0].par_env
    #     except Exception:
    #         pass
    if isinstance(env, VecEnv):

        scenario: sim.Scenario | None = env.get_attr('_scenario')[0]
        max_duration: float = env.get_attr('max_duration')[0]
        time_step: float = env.get_attr('time_step')[0]
        env_groups = env.get_attr('groups_config')[0]
        possible_agents = env.get_attr('_possible_agents')[0]
    else:
        if not isinstance(env.unwrapped, NavgroundBaseEnv):
            raise TypeError(
                f"Environment {env} is not a Navground environment.")
        scenario = env.unwrapped._scenario
        max_duration = env.unwrapped.max_duration
        time_step = env.unwrapped.time_step
        env_groups = env.unwrapped.groups_config
        possible_agents = env.unwrapped._possible_agents

    experiment = sim.Experiment()
    experiment.terminate_when_all_idle_or_stuck = False
    if scenario:
        experiment.scenario = copy.deepcopy(scenario)
    if max_duration > 0:
        experiment.steps = int(max_duration / time_step)
    init = InitPolicyBehavior.with_env(env=env,
                                       groups=groups,
                                       deterministic=deterministic,
                                       grouped=grouped,
                                       pre=pre)
    experiment.scenario.add_init(init)
    groups = merge_groups_configs(groups, env_groups, len(possible_agents))
    if record_success:
        indices = join_indices(
            (group.indices for group in groups
             if group.success_condition or group.failure_condition),
            len(possible_agents))
        if indices:
            experiment.add_record_probe("success",
                                        SuccessProbe.with_indices(indices))
    if record_reward:
        experiment.add_record_probe(
            "reward", partial(RewardProbe, groups=groups, reward=reward))
    return experiment



from __future__ import annotations

import gymnasium as gym
import numpy as np

from ..types import PathLike, PyTorchPolicy


def export(policy: PyTorchPolicy,
           path: PathLike,
           dynamic_batch_size: bool = True) -> None:
    """
    Export a policy (PyTorch) as onnx.

    :param      policy:              The policy
    :param      path:                Where to save the policy;
        by convention should have an ".onnx" suffix.
    :param      dynamic_batch_size:  Whether to enable dynamic batch size.
        If not set, it will export a model that evaluate a single sample.
    """
    import torch as th

    from .onnxable import OnnxablePolicy, OnnxablePolicyWithMultiInput

    # if hasattr(policy, 'actor'):
    #     policy = policy.actor
    obs = policy.observation_space.sample()
    if isinstance(obs, dict):
        dummy_input = tuple(
            th.from_numpy(v[np.newaxis, :]) for v in obs.values())
    else:
        dummy_input = (th.from_numpy(obs[np.newaxis, :]), )
    onnxable_model: th.nn.Module
    if isinstance(policy.observation_space, gym.spaces.Dict):
        keys = list(policy.observation_space.keys())
        onnxable_model = OnnxablePolicyWithMultiInput(policy, keys)
        input_names = keys
    else:
        onnxable_model = OnnxablePolicy(policy)
        input_names = ["observation"]

    if dynamic_batch_size:
        dynamic_axes: dict[str, dict[int, str]] | None = {
            k: {
                0: 'batch_size'
            }
            for k in input_names
        }
    else:
        dynamic_axes = None
    th.onnx.export(onnxable_model,
                   dummy_input,
                   path,
                   opset_version=17,
                   input_names=input_names,
                   dynamic_axes=dynamic_axes,
                   output_names=["action"])



from __future__ import annotations

import yaml
from navground import core
from collections.abc import Collection


class SelectionBehavior(core.Behavior, name="Selection"):

    def __init__(self,
                 kinematics: core.Kinematics | None = None,
                 radius: float = 0.0,
                 behaviors: Collection[core.Behavior] = tuple()):
        super().__init__(kinematics, radius)
        self.behaviors = list(behaviors)
        self._state = core.GeometricState()
        self._index = 0

    def decode(self, value: str) -> None:
        data = yaml.safe_load(value)
        behaviors = data.get('behaviors', [])
        for behavior in behaviors:
            b = core.load_behavior(yaml.safe_dump(behavior))
            if b:
                self.behaviors.append(b)

    def encode(self) -> str:
        behaviors = [
            yaml.safe_load(core.dump(behavior)) for behavior in self.behaviors
        ]
        if behaviors:
            return yaml.safe_dump({'behaviors': behaviors})
        return ''

    @property
    @core.register(0, "The index of the active behavior")
    def index(self) -> int:
        return self._index

    @index.setter
    def index(self, value: int) -> None:
        self._index = max(0, value)

    def compute_cmd_internal(self, time_step: float) -> core.Twist2:
        if self.index < len(self.behaviors):
            behavior = self.behaviors[self.index]
            behavior.set_state_from(self)
            state = behavior.environment_state
            if isinstance(state, core.GeometricState):
                state.neighbors = self._state.neighbors
                state.static_obstacles = self._state.static_obstacles
                state.line_obstacles = self._state.line_obstacles
            return behavior.compute_cmd_internal(time_step)
        return super().compute_cmd_internal(time_step)

    def get_environment_state(self) -> core.EnvironmentState:
        return self._state
